from __future__ import annotations

import copy
import html
import uuid
import warnings
from itertools import product
from typing import TYPE_CHECKING, Literal, cast

import numpy as np
import skrub
from joblib import Parallel
from numpy.typing import ArrayLike
from sklearn.base import BaseEstimator, clone
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline
from sklearn.utils._response import (
    _check_response_method,
    _process_decision_function,
    _process_predict_proba,
)
from sklearn.utils.validation import check_is_fitted

from skore._externals._pandas_accessors import DirNamesMixin
from skore._externals._sklearn_compat import is_clusterer
from skore._sklearn._base import _BaseReport
from skore._sklearn.find_ml_task import _find_ml_task
from skore._sklearn.types import DataSource, PositiveLabel
from skore._utils._cache import Cache
from skore._utils._cache_key import make_cache_key
from skore._utils._fixes import _validate_joblib_parallel_params
from skore._utils._measure_time import MeasureTime
from skore._utils._parallel import delayed
from skore._utils._progress_bar import track
from skore._utils.repr.data import get_documentation_url
from skore._utils.repr.html_repr import render_template

if TYPE_CHECKING:
    from skore._sklearn._estimator.data_accessor import _DataAccessor
    from skore._sklearn._estimator.inspection_accessor import (
        _InspectionAccessor,
    )
    from skore._sklearn._estimator.metrics_accessor import _MetricsAccessor


def _compute_predictions(estimator, X, response_method):
    prediction_method = _check_response_method(estimator, response_method)
    with MeasureTime() as predict_time:
        predictions = prediction_method(X)

    return predictions, predict_time()


class EstimatorReport(_BaseReport, DirNamesMixin):
    """Report for a fitted estimator.

    This class provides a set of tools to quickly validate and inspect a scikit-learn
    compatible estimator.

    Refer to the :ref:`estimator_report` section of the user guide for more details.

    Parameters
    ----------
    estimator : estimator object
        Estimator to make the report from. When the estimator is not fitted,
        it is deep-copied to avoid side-effects. If it is fitted, it is cloned instead.

    fit : {"auto", True, False}, default="auto"
        Whether to fit the estimator on the training data. If "auto", the estimator
        is fitted only if the training data is provided.

    X_train : {array-like, sparse matrix} of shape (n_samples, n_features) or \
            None
        Training data.

    y_train : array-like of shape (n_samples,) or (n_samples, n_outputs) or None
        Training target.

    X_test : {array-like, sparse matrix} of shape (n_samples, n_features)
        Testing data. It should have the same structure as the training data.

    y_test : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Testing target.

    pos_label : int, float, bool or str, default=None
        For binary classification, the positive class to use for metrics and displays
        that need one. If `None`, skore does not infer a default positive class.
        Binary metrics and displays that support it will expose all classes instead.
        This parameter is rejected for non-binary tasks.

    Attributes
    ----------
    estimator_ : estimator object
        The cloned or copied estimator.

    estimator_name_ : str
        The name of the estimator.

    fit_time_ : float or None
        The time taken to fit the estimator, in seconds. If the estimator is not
        internally fitted, the value is `None`.

    See Also
    --------
    skore.CrossValidationReport
        Report of cross-validation results.

    skore.ComparisonReport
        Report of comparison between estimators.

    Examples
    --------
    >>> from sklearn.datasets import make_classification
    >>> from skore import train_test_split
    >>> from sklearn.linear_model import LogisticRegression
    >>> X, y = make_classification(random_state=42)
    >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
    >>> estimator = LogisticRegression()
    >>> from skore import EstimatorReport
    >>> report = EstimatorReport(estimator, **split_data)
    """

    _ACCESSOR_CONFIG: dict[str, dict[str, str]] = {
        "data": {"name": "data"},
        "metrics": {"name": "metrics"},
        "inspection": {"name": "inspection"},
    }

    _report_type: Literal["estimator"] = "estimator"

    metrics: _MetricsAccessor
    inspection: _InspectionAccessor
    data: _DataAccessor

    @staticmethod
    def _fit_estimator(
        estimator: BaseEstimator,
        X_train: ArrayLike | None,
        y_train: ArrayLike | None,
    ) -> tuple[BaseEstimator, float]:
        if X_train is None or y_train is None:
            raise ValueError(
                "The training data is required to fit the estimator. "
                "Please provide both X_train and y_train."
            )
        estimator_ = clone(estimator)
        with MeasureTime() as fit_time:
            estimator_.fit(X_train, y_train)
        return estimator_, fit_time()

    @classmethod
    def _copy_estimator(cls, estimator: BaseEstimator) -> BaseEstimator:
        try:
            return copy.deepcopy(estimator)
        except Exception as e:
            warnings.warn(
                "Deepcopy failed; using estimator as-is. "
                "Be aware that modifying the estimator outside of "
                f"{cls.__name__} will modify the internal estimator. "
                "Consider using a FrozenEstimator from scikit-learn to prevent this. "
                f"Original error: {e}",
                stacklevel=1,
            )
        return estimator

    def __init__(
        self,
        estimator: BaseEstimator,
        *,
        fit: Literal["auto"] | bool = "auto",
        X_train: ArrayLike | None = None,
        y_train: ArrayLike | None = None,
        X_test: ArrayLike,
        y_test: ArrayLike,
        pos_label: PositiveLabel | None = None,
    ) -> None:
        super().__init__()
        self._fit = fit

        if is_clusterer(estimator):
            raise ValueError(
                "Clustering models are not supported yet. Please use a"
                " classification or regression model instead."
            )

        fit_time: float | None = None
        if fit == "auto":
            try:
                check_is_fitted(estimator)
                self._estimator = self._copy_estimator(estimator)
            except NotFittedError:
                self._estimator, fit_time = self._fit_estimator(
                    estimator, X_train, y_train
                )
        elif fit is True:
            self._estimator, fit_time = self._fit_estimator(estimator, X_train, y_train)
        else:  # fit is False
            self._estimator = self._copy_estimator(estimator)

        # private storage to ensure properties are read-only
        self._X_train = X_train
        self._y_train = y_train
        self._X_test = X_test
        self._y_test = y_test
        self._pos_label = pos_label
        self.fit_time_ = fit_time
        self._cache = Cache()

        self._ml_task = _find_ml_task(self._y_test, estimator=self._estimator)
        if pos_label is None:
            return
        if self._ml_task != "binary-classification":
            raise ValueError(
                "pos_label is only accepted/used for binary classification"
            )
        labels = self.estimator_.classes_.tolist()
        if pos_label not in labels:
            raise ValueError(
                f"pos_label={pos_label!r} is not a valid label. "
                f"It should be one of: {labels!r}."
            )

        # NOTE: Reports are immutable so we don't need cache invalidation

    def clear_cache(self) -> None:
        """Clear the cache.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> report = EstimatorReport(classifier, **split_data)
        >>> report.cache_predictions()
        >>> report.clear_cache()
        >>> report._cache
        {}
        """
        self._cache = Cache()

    def cache_predictions(
        self,
        response_methods: Literal["auto"] | str | list[str] = "auto",
        data_source: DataSource | Literal["both"] = "both",
        n_jobs: int | None = None,
    ) -> None:
        """Cache estimator's predictions.

        Parameters
        ----------
        response_methods : "auto" or list of str, default="auto"
            The response methods to precompute. If "auto", the response methods are
            inferred from the ml task: for classification we compute the response of
            the `predict_proba`, `decision_function` and `predict` methods; for
            regression we compute the response of the `predict` method.

        data_source : {"test", "train", "both"}, default="both"
            The data source(s) for which to precompute predictions.

            - "test" : cache predictions for the test set only.
            - "train" : cache predictions for the train set only.
            - "both" : cache predictions for both train and test sets when available.

        n_jobs : int or None, default=None
            The number of jobs to run in parallel. None means 1 unless in a
            joblib.parallel_backend context. -1 means using all processors.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> report = EstimatorReport(classifier, **split_data)
        >>> report.cache_predictions()
        >>> report._cache
        {...}
        """
        if self._ml_task in ("binary-classification", "multiclass-classification"):
            if response_methods == "auto":
                response_methods = ["predict"]
                if hasattr(self._estimator, "predict_proba"):
                    response_methods += ["predict_proba"]
                if hasattr(self._estimator, "decision_function"):
                    response_methods += ["decision_function"]
        else:
            if response_methods == "auto":
                response_methods = ["predict"]

        data_sources = []
        if data_source in ("test", "both"):
            data_sources += [("test", self._X_test)]
        if data_source in ("train", "both") and self._X_train is not None:
            data_sources += [("train", self._X_train)]

        keys_and_to_compute = [
            (
                (cast(DataSource, ds), response_method),
                (self._estimator, X, response_method),
            )
            for response_method, (ds, X) in product(response_methods, data_sources)
        ]

        keys_and_to_compute = [
            (key, data)
            for key, data in keys_and_to_compute
            if make_cache_key(*key) not in self._cache
        ]

        if not keys_and_to_compute:
            return

        parallel = Parallel(
            **_validate_joblib_parallel_params(n_jobs=n_jobs, return_as="generator")
        )

        # trigger the computation
        # do not mutate directly `self._cache` during the execution of Parallel
        keys, to_compute = zip(*keys_and_to_compute, strict=True)

        for (ds, response_method), (preds, time) in zip(
            keys,
            track(
                parallel(delayed(_compute_predictions)(*args) for args in to_compute),
                description="Caching predictions",
                total=(len(response_methods) * len(data_sources)),
            ),
            strict=True,
        ):
            self._cache[make_cache_key(ds, response_method)] = preds
            if response_method == "predict":
                self._cache[make_cache_key(ds, "predict_time")] = time

    def get_predictions(
        self,
        *,
        data_source: Literal["train", "test"],
        response_method: Literal[
            "predict", "predict_proba", "decision_function"
        ] = "predict",
    ) -> ArrayLike:
        """Get estimator's predictions.

        This method has the advantage to reload from the cache if the predictions
        were already computed in a previous call.

        Parameters
        ----------
        data_source : {"test", "train"}, default="test"
            The data source to use.

            - "test" : use the test set provided when creating the report.
            - "train" : use the train set provided when creating the report.

        response_method : {"predict", "predict_proba", "decision_function"}, \
                default="predict"
            The response method to use to get the predictions.

        Returns
        -------
        np.ndarray of shape (n_samples,) or (n_samples, n_classes)
            The predictions.

        Raises
        ------
        ValueError
            If the data source is invalid.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from skore import train_test_split
        >>> from sklearn.linear_model import LogisticRegression
        >>> X, y = make_classification(random_state=42)
        >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
        >>> estimator = LogisticRegression()
        >>> from skore import EstimatorReport
        >>> report = EstimatorReport(estimator, **split_data)
        >>> predictions = report.get_predictions(data_source="test")
        >>> predictions.shape
        (25,)
        """
        pos_label = self.pos_label
        if (
            pos_label is None
            and self.ml_task == "binary-classification"
            and response_method == "decision_function"
        ):
            # we do this to follow scikit-learn convention:
            pos_label = self.estimator_.classes_[-1]
        return self._get_predictions(
            data_source=data_source,
            response_method=response_method,
            pos_label=pos_label,
        )

    def _get_predictions(
        self,
        *,
        data_source: Literal["train", "test"],
        response_method: str | list[str] | tuple[str, ...],
        pos_label: PositiveLabel | None = None,
    ) -> ArrayLike:
        """Get estimator's predictions, and adapt them to `pos_label` if needed.

        Internal helpers used by the metrics.

        Returns
        -------
        np.ndarray of shape (n_samples,) or (n_samples, n_classes)
            The predictions.
            The shape is (n_samples,) if:
                - response_method is "predict"
                - OR if pos_label is specified (binary-classification only)
            Otherwise it's (n_samples, n_classes)
        """
        if data_source not in ("train", "test"):
            raise ValueError(f"Invalid data source: {data_source}")
        if pos_label is not None and self.ml_task != "binary-classification":
            raise ValueError(f"Cannot specify a `pos_label` for task {self.ml_task}")

        method_name = _check_response_method(self.estimator_, response_method).__name__
        self.cache_predictions(response_methods=[method_name], data_source=data_source)
        cache_key = make_cache_key(data_source, method_name)
        predictions = self._cache[cache_key]

        if method_name == "predict":
            return predictions
        # Adapt to pos_label / check shape if needed:
        if method_name in ("predict_proba", "predict_log_proba"):
            if pos_label is None:
                return predictions
            # Adapt to pos_label:
            return _process_predict_proba(
                y_pred=predictions,
                target_type="binary",
                classes=self.estimator_.classes_,
                pos_label=pos_label,
            )
        elif method_name == "decision_function":
            if pos_label is not None:
                # Adapt to pos_label:
                return _process_decision_function(
                    y_pred=predictions,
                    target_type="binary",
                    classes=self.estimator_.classes_,
                    pos_label=pos_label,
                )

            if self.ml_task == "multiclass-classification":
                _, d = predictions.shape
                if d != len(self.estimator_.classes_):
                    raise ValueError(f"Unexpected decision function shape[1]: {d}")
                return predictions
            # scikit-learn returns a (n_samples,) array that corresponds to classes[-1]
            # we normalize to a (n_samples, 2) shape with similar semantic than
            # predict_proba
            return np.vstack((-predictions, predictions)).T
        else:
            raise ValueError(f"Unexpected response_method: {method_name}")

    @property
    def ml_task(self):
        return self._ml_task

    @property
    def estimator(self) -> BaseEstimator:
        return self._estimator

    @property
    def estimator_(self) -> BaseEstimator:
        return self._estimator

    @property
    def X_train(self) -> ArrayLike | None:
        return self._X_train

    @property
    def y_train(self) -> ArrayLike | None:
        return self._y_train

    @property
    def X_test(self) -> ArrayLike | None:
        return self._X_test

    @property
    def y_test(self) -> ArrayLike | None:
        return self._y_test

    @property
    def pos_label(self) -> PositiveLabel | None:
        return self._pos_label

    @property
    def estimator_name_(self) -> str:
        if isinstance(self._estimator, Pipeline):
            name = self._estimator[-1].__class__.__name__
        else:
            name = self._estimator.__class__.__name__
        return name

    @property
    def fit(self) -> str | bool:
        return self._fit

    ####################################################################################
    # Methods related to the help and repr
    ####################################################################################

    def _get_help_title(self) -> str:
        return f"Tools to diagnose estimator {self.estimator_name_}"

    def _get_help_legend(self) -> str:
        return (
            "[cyan](↗︎)[/cyan] higher is better [orange1](↘︎)[/orange1] lower is better"
        )

    def __repr__(self) -> str:
        """Return a string representation."""
        return f"""{self.__class__.__name__}:
        {self.estimator_!r}

        {self.metrics.summarize().frame()}"""

    def _html_repr_fragments(self) -> dict[str, str]:
        """HTML snippets for the report body (metrics, estimator diagram, data table).

        Used by :meth:`_repr_html_` and by :class:`~skore.ComparisonReport` to embed
        one report's views in the comparison HTML repr.
        """
        match self.X_train, self.X_test:
            case None, None:
                data_source = None
            case _, None:
                data_source = "train"
            case None, _:
                data_source = "test"
            case _:
                data_source = "both"
        if data_source is None:
            table_report_html = "<p>No data provided</p>"
            metrics_html = "<p>No data provided</p>"
        else:
            table_report = skrub.TableReport(
                self.data._prepare_dataframe_for_display(
                    data_source=data_source,
                    with_y=True,
                    subsample=None,
                    subsample_strategy="head",
                    seed=None,
                ),
                max_plot_columns=0,
                max_association_columns=0,
                verbose=False,
            )
            table_report._set_minimal_mode()
            table_report_html = table_report.html_snippet()
            metrics_html = (
                self.metrics.summarize(
                    data_source="train" if data_source == "train" else "test"
                )
                .frame()
                ._repr_html_()
            )
        try:
            estimator_html = self.estimator_._repr_html_()
        except Exception:
            estimator_html = f"<p>{html.escape(repr(self.estimator_))}</p>"

        return {
            "metrics_summary": metrics_html,
            "estimator_display": estimator_html,
            "table_report": table_report_html,
        }

    def _repr_html_(self) -> str:
        """HTML representation of estimator.

        This is redundant with the logic of `_repr_mimebundle_`. The latter
        should be favored in the long term, `_repr_html_` is only
        implemented for consumers who do not interpret `_repr_mimbundle_`.
        """
        fragments = self._html_repr_fragments()
        container_id = f"skore-estimator-report-{uuid.uuid4().hex[:8]}"
        help_doc_url = get_documentation_url(obj=self, method_name="help")
        report_class_name = self.__class__.__name__
        metrics_accessor_doc_url = get_documentation_url(
            obj=self, accessor_name="metrics"
        )
        inspection_accessor_doc_url = get_documentation_url(
            obj=self, accessor_name="inspection"
        )
        data_accessor_doc_url = get_documentation_url(obj=self, accessor_name="data")
        return render_template(
            "estimator_report.html.j2",
            {
                "container_id": container_id,
                "help_doc_url": help_doc_url,
                "report_class_name": report_class_name,
                "metrics_accessor_doc_url": metrics_accessor_doc_url,
                "inspection_accessor_doc_url": inspection_accessor_doc_url,
                "data_accessor_doc_url": data_accessor_doc_url,
                **fragments,
            },
        )

    def _repr_mimebundle_(self, **kwargs):
        """Mime bundle used by jupyter kernels to display estimator."""
        output = {"text/plain": repr(self)}
        output["text/html"] = self._repr_html_()
        return output



from collections.abc import Sequence
from typing import Any, Literal, cast

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from numpy.typing import NDArray
from pandas import CategoricalDtype, DataFrame
from sklearn.preprocessing import LabelBinarizer

from skore._sklearn.types import (
    DataSource,
    MLTask,
    PositiveLabel,
    ReportType,
)

LINESTYLE = [
    ("solid", "solid"),
    ("dotted", "dotted"),
    ("dashed", "dashed"),
    ("dashdot", "dashdot"),
    ("loosely dotted", (0, (1, 10))),
    ("dotted", (0, (1, 5))),
    ("densely dotted", (0, (1, 1))),
    ("long dash with offset", (5, (10, 3))),
    ("loosely dashed", (0, (5, 10))),
    ("dashed", (0, (5, 5))),
    ("densely dashed", (0, (5, 1))),
    ("loosely dashdotted", (0, (3, 10, 1, 10))),
    ("dashdotted", (0, (3, 5, 1, 5))),
    ("densely dashdotted", (0, (3, 1, 1, 1))),
    ("dashdotdotted", (0, (3, 5, 1, 5, 1, 5))),
    ("loosely dashdotdotted", (0, (3, 10, 1, 10, 1, 10))),
    ("densely dashdotdotted", (0, (3, 1, 1, 1, 1, 1))),
]


class _ClassifierDisplayMixin:
    """Mixin class to be used in Displays requiring a classifier."""

    # defined in the concrete display class
    estimator_name: str
    ml_task: MLTask
    pos_label: PositiveLabel | None


def _rotate_ticklabels(
    ax: Axes, *, rotation: float = 45, horizontalalignment: str = "right"
) -> None:
    """Rotate the tick labels of the axis.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axis to rotate.
    rotation : float, default=45
        The rotation of the tick labels.
    horizontalalignment : {"left", "center", "right"}, default="right"
        The horizontal alignment of the tick labels.
    """
    plt.setp(
        ax.get_xticklabels(), rotation=rotation, horizontalalignment=horizontalalignment
    )


def _get_adjusted_fig_size(
    fig: Figure, ax: Axes, direction: str, target_size: float
) -> float:
    """Get the adjusted figure size.

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        The figure to adjust.

    ax : matplotlib.axes.Axes
        The axis to adjust.

    direction : {"width", "height"}
        The direction to adjust.

    target_size : float
        The target size.

    Returns
    -------
    float
        The adjusted figure size.
    """
    size_display = getattr(ax.get_window_extent(), direction)
    size = fig.dpi_scale_trans.inverted().transform((size_display, 0))[0]
    dim = 0 if direction == "width" else 1
    fig_size = fig.get_size_inches()[dim]
    return target_size * (fig_size / size)


def _adjust_fig_size(
    fig: Figure, ax: Axes, target_width: float, target_height: float
) -> None:
    """Rescale a figure to the target width and height.

    This allows us to have all figures of a given type (bar plots, lines or
    histograms) have the same size, so that the displayed report looks more
    uniform, without having to do manual adjustments to account for the length
    of labels, occurrence of titles etc. We let pyplot generate the figure
    without any size constraints then resize it and thus let matplotlib deal
    with resizing the appropriate elements (eg shorter bars when the labels
    take more horizontal space).

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        The figure to adjust.

    ax : matplotlib.axes.Axes
        The axis to adjust.

    target_width : float
        The target width.

    target_height : float
        The target height.
    """
    width = _get_adjusted_fig_size(fig, ax, "width", target_width)
    height = _get_adjusted_fig_size(fig, ax, "height", target_height)
    fig.set_size_inches((width, height))


def _despine_matplotlib_axis(
    ax: Axes,
    *,
    axis_to_despine: tuple = ("top", "right"),
    remove_ticks: bool = False,
    x_range: tuple[float, float] | None = (0, 1),
    y_range: tuple[float, float] | None = (0, 1),
    offset: float = 10,
) -> None:
    """Despine the matplotlib axis.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The matplotlib axis to despine.
    axis_to_despine : tuple of str, default=("top", "right")
        The axes to despine.
    remove_ticks : bool, default=False
        Whether to remove the ticks when adding "left" and "bottom" to the axis to
        despine.
    x_range : tuple of float, default=(0, 1)
        The range of the x-axis.
    y_range : tuple of float, default=(0, 1)
        The range of the y-axis.
    offset : float, default=10
        The offset to add to the x-axis and y-axis.
    """
    if x_range is not None:
        ax.spines["bottom"].set_bounds(x_range[0], x_range[1])
    if y_range is not None:
        ax.spines["left"].set_bounds(y_range[0], y_range[1])
    ax.spines["left"].set_position(("outward", offset))
    ax.spines["bottom"].set_position(("outward", offset))
    for s in axis_to_despine:
        ax.spines[s].set_visible(False)
        if remove_ticks and s in ("left", "bottom"):
            axis: Literal["x", "y"] = "x" if s == "bottom" else "y"
            ax.tick_params(axis=axis, length=0)


def _validate_style_kwargs(
    default_style_kwargs: dict[str, Any], user_style_kwargs: dict[str, Any]
) -> dict[str, Any]:
    """Create valid style kwargs by avoiding Matplotlib alias errors.

    Matplotlib raises an error when, for example, 'color' and 'c', or 'linestyle' and
    'ls', are specified together. To avoid this, we automatically keep only the one
    specified by the user and raise an error if the user specifies both.

    Parameters
    ----------
    default_style_kwargs : dict
        The Matplotlib style kwargs used by default in the scikit-learn display.
    user_style_kwargs : dict
        The user-defined Matplotlib style kwargs.

    Returns
    -------
    valid_style_kwargs : dict
        The validated style kwargs taking into account both default and user-defined
        Matplotlib style kwargs.
    """
    invalid_to_valid_kw = {
        "ls": "linestyle",
        "c": "color",
        "ec": "edgecolor",
        "fc": "facecolor",
        "lw": "linewidth",
        "mec": "markeredgecolor",
        "mfcalt": "markerfacecoloralt",
        "ms": "markersize",
        "mew": "markeredgewidth",
        "mfc": "markerfacecolor",
        "aa": "antialiased",
        "ds": "drawstyle",
        "font": "fontproperties",
        "family": "fontfamily",
        "name": "fontname",
        "size": "fontsize",
        "stretch": "fontstretch",
        "style": "fontstyle",
        "variant": "fontvariant",
        "weight": "fontweight",
        "ha": "horizontalalignment",
        "va": "verticalalignment",
        "ma": "multialignment",
    }
    for invalid_key, valid_key in invalid_to_valid_kw.items():
        if invalid_key in user_style_kwargs and valid_key in user_style_kwargs:
            raise TypeError(
                f"Got both {invalid_key} and {valid_key}, which are aliases of one "
                "another"
            )
    valid_style_kwargs = default_style_kwargs.copy()

    for key in user_style_kwargs:
        if key in invalid_to_valid_kw:
            valid_style_kwargs[invalid_to_valid_kw[key]] = user_style_kwargs[key]
        else:
            valid_style_kwargs[key] = user_style_kwargs[key]

    return valid_style_kwargs


def sample_mpl_colormap(
    cmap: Colormap, n: int
) -> list[tuple[float, float, float, float]]:
    """Sample colors from a Matplotlib colormap.

    Parameters
    ----------
    cmap : matplotlib.colors.Colormap
        The Matplotlib colormap to sample from.
    n : int
        The number of colors to sample.

    Returns
    -------
    colors : list of str
        The sampled colors.
    """
    indices = np.linspace(0, 1, n)
    return [cmap(i) for i in indices]


def _get_curve_plot_columns(
    plot_data: DataFrame,
    report_type: ReportType,
    pos_label,
    data_source: DataSource | Literal["both"],
    subplot_by: Literal["auto", "label", "estimator", "data_source"] | None = "auto",
) -> tuple[str | None, str | None, str | None]:
    """Determine col, hue, and style columns for precision-recall and ROC.

    Rules:
    - Default ("auto"): None for EstimatorReport and Cross-Validation Report,
        "estimator" for ComparisonReport
    - subplot_by=None disallowed for comparison when plotting one-vs-rest curves
    - subplot_by="estimator" only allowed for comparison reports
    - subplot_by="label" only allowed when plotting one-vs-rest curves
    - subplot_by="data_source" only allowed for EstimatorReport with both data \
        sources
    - hue priority: estimator > label > data_source (excluding col)

    Returns (col, hue, style) tuple where each can be None if not applicable.
    """
    has_multiple_estimators = (
        "estimator" in plot_data.columns and plot_data["estimator"].nunique() > 1
    )
    is_comparison = "comparison" in report_type
    is_one_vs_rest = pos_label is None
    has_both_data_sources = data_source == "both"

    allowed_values: set[str | None] = {"auto"}
    if is_one_vs_rest:
        allowed_values.add("label")
        if not is_comparison:
            allowed_values.add(None)
    else:
        allowed_values.add(None)
    if is_comparison and has_multiple_estimators:
        allowed_values.add("estimator")
    if has_both_data_sources and (not is_comparison or not is_one_vs_rest):
        allowed_values.add("data_source")
    # Disallow for comparison reports when plotting one-vs-rest curves.

    if subplot_by not in allowed_values:
        allowed_str = ", ".join(sorted([str(s) for s in allowed_values]))
        raise ValueError(
            f"subplot_by must be one of {allowed_str}. Got {subplot_by!r} instead."
        )

    if subplot_by == "auto":
        col = "estimator" if is_comparison else None
    else:
        col = subplot_by
    has_multiple_labels = (
        "label" in plot_data.columns and plot_data["label"].nunique() > 1
    )

    hue_candidates = []
    if has_multiple_estimators:
        hue_candidates.append("estimator")
    if has_multiple_labels:
        hue_candidates.append("label")

    hue = hue[0] if (hue := [c for c in hue_candidates if c != col]) else None

    style = "data_source" if has_both_data_sources else None

    return col, hue, style


def _build_custom_legend_with_stats(
    *,
    ax: Axes,
    subplot_data: DataFrame,
    hue: str | None,
    style: str | None,
    hue_order: list[Any] | None,
    style_order: list[Any] | None,
    is_cross_validation: bool = False,
    statistic_column_name: Literal["average_precision", "roc_auc"],
    statistic_acronym: Literal["AP", "AUC"],
    chance_level_label: str | None = None,
) -> None:
    """Build custom legend with a custom statistic for a single axis."""
    legend_labels = []
    for hue_value in hue_order or [None]:
        hue_value_str = (
            f"'{hue_value}'" if isinstance(hue_value, str) else str(hue_value)
        )
        hue_group = (
            subplot_data.query(f"{hue} == {hue_value_str}")
            if hue_value is not None
            else subplot_data
        )
        for style_value in style_order or [None]:
            style_value_str = (
                f"'{style_value}'" if isinstance(style_value, str) else str(style_value)
            )
            statistic_group = (
                hue_group.query(f"{style} == {style_value_str}")[statistic_column_name]
                if style_value is not None
                else hue_group[statistic_column_name]
            )
            if not statistic_group.empty:
                statistic = (
                    f"{statistic_acronym}={statistic_group.mean():.2f}±{statistic_group.std():.2f}"
                    if is_cross_validation and len(statistic_group) > 1
                    else f"{statistic_acronym}={statistic_group.iloc[0]:.2f}"
                )
                legend_labels.append(
                    _format_legend_label(
                        style_column_name=style,
                        style_value=style_value,
                        hue_value=hue_value,
                        statistic=statistic,
                    )
                )
    if chance_level_label is not None:
        legend_labels.append(chance_level_label)

    n_entries = len(legend_labels)
    lines = ax.get_lines()
    handles = []
    seen_line_attributes = []
    for line in lines:
        line_attributes = (line.get_color(), line.get_linestyle())
        if line_attributes not in seen_line_attributes:
            seen_line_attributes.append(line_attributes)
            handles.append(line)

    fontsize = "small" if n_entries > 4 else "medium"

    ax.legend(
        handles,
        legend_labels,
        loc="upper center",
        bbox_to_anchor=(0.5, -0.15),
        ncol=1,
        frameon=False,
        fontsize=fontsize,
    )


def _format_legend_label(
    style_column_name: str | None,
    style_value: str | None,
    hue_value: str | None,
    statistic: str,
) -> str:
    """Format a legend label based on style and hue."""
    if style_value is None and hue_value is None:
        return statistic
    if style_column_name == "data_source":
        style_value = cast(str, style_value)
        style_value = style_value.title() + " set"
    return (
        " - ".join(
            str(s) for s in filter(lambda x: x is not None, [hue_value, style_value])
        )
        + f" ({statistic})"
    )


def _column_data_to_records(column_data: dict[str, list] | None) -> list[dict]:
    """Convert column-wise data into one record per row.

    Examples
    --------
    >>> _column_data_to_records({"estimator": ["A", "B"], "split": [0, 1]})
    [{'estimator': 'A', 'split': 0}, {'estimator': 'B', 'split': 1}]
    """
    if column_data is None or not column_data:
        return []

    keys = list(column_data)
    values = [list(column_data[key]) for key in keys]
    lengths = {len(value) for value in values}
    if len(lengths) > 1:
        raise ValueError("All column data iterables must have the same length.")

    return [dict(zip(keys, row, strict=True)) for row in zip(*values, strict=True)]


def _concat_frames_with_column_data(
    frames: Sequence[DataFrame], column_data: dict[str, list] | None = None
) -> DataFrame:
    """Concatenate frames and assign column-wise metadata as categorical columns."""
    column_records = _column_data_to_records(column_data)
    if not column_records:
        column_records = [{} for _ in frames]

    frame = pd.concat(
        [
            child_frame.assign(**column_record)
            for child_frame, column_record in zip(frames, column_records, strict=True)
        ],
        ignore_index=True,
    )

    categorical_columns = {
        column
        for column in frame.columns
        if all(
            column in child_frame.columns
            and isinstance(child_frame[column].dtype, CategoricalDtype)
            for child_frame in frames
        )
    }
    if column_data:
        categorical_columns.update(column_data)

    if categorical_columns:
        frame = frame.astype(dict.fromkeys(categorical_columns, "category"))

    return frame


def _one_hot_encode(y_true, classes) -> NDArray:
    label_binarizer = LabelBinarizer().fit(classes)
    y_true_onehot: NDArray = label_binarizer.transform(y_true)
    if len(classes) == 2:
        y_true_onehot = np.hstack(((1 - y_true_onehot), y_true_onehot))
    return y_true_onehot



from __future__ import annotations

from sklearn.utils.metaestimators import available_if

from skore._externals._pandas_accessors import DirNamesMixin
from skore._sklearn._base import _BaseAccessor
from skore._sklearn._estimator.report import EstimatorReport
from skore._sklearn._plot.inspection.coefficients import CoefficientsDisplay
from skore._sklearn._plot.inspection.impurity_decrease import ImpurityDecreaseDisplay
from skore._sklearn._plot.inspection.permutation_importance import (
    PermutationImportanceDisplay,
)
from skore._sklearn.types import DataSource, Metric
from skore._utils._accessor import (
    _check_estimator_has_coef,
    _check_estimator_has_feature_importances,
)
from skore._utils._cache_key import make_cache_key


class _InspectionAccessor(_BaseAccessor[EstimatorReport], DirNamesMixin):
    """Accessor for model inspection related operations.

    You can access this accessor using the `inspection` attribute.
    """

    def __init__(self, parent: EstimatorReport) -> None:
        super().__init__(parent)

    @available_if(_check_estimator_has_coef())
    def coefficients(self) -> CoefficientsDisplay:
        """Retrieve the coefficients of a linear model, including the intercept.

        Returns
        -------
        :class:`CoefficientsDisplay`
            The feature importance display containing model coefficients and
            intercept.

        Examples
        --------
        >>> from sklearn.datasets import load_diabetes
        >>> from sklearn.linear_model import Ridge
        >>> from skore import evaluate
        >>> X, y = load_diabetes(return_X_y=True)
        >>> regressor = Ridge()
        >>> report = evaluate(regressor, X, y, splitter=0.2)
        >>> display = report.inspection.coefficients()
        >>> display.frame()
               feature  coefficient
        0    Intercept      151.9...
        1   Feature #0       21.3...
        2   Feature #1      -72.9...
        3   Feature #2      301.3...
        4   Feature #3      177.4...
        5   Feature #4        2.8...
        6   Feature #5      -35.2...
        7   Feature #6     -155.5...
        8   Feature #7      118.3...
        9   Feature #8      257.3...
        10  Feature #9      102.2...
        >>> display.plot() # shows plot
        """
        return CoefficientsDisplay._compute_data_for_display(
            estimator=self._parent.estimator_,
            name=self._parent.estimator_name_,
            report_type=self._parent._report_type,
        )

    @available_if(_check_estimator_has_feature_importances())
    def impurity_decrease(self) -> ImpurityDecreaseDisplay:
        """Retrieve the Mean Decrease in Impurity (MDI) of a tree-based model.

        This method is available for estimators that expose a `feature_importances_`
        attribute. See for example
        :attr:`sklearn.ensemble.GradientBoostingClassifier.feature_importances_`.

        In particular, note that the MDI is computed at fit time, i.e. using the
        training data.

        Returns
        -------
        :class:`ImpurityDecreaseDisplay`
            The feature importance display containing the mean decrease in impurity.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from skore import evaluate
        >>> X, y = make_classification(n_features=5, random_state=42)
        >>> forest = RandomForestClassifier(n_estimators=5, random_state=0)
        >>> report = evaluate(forest, X, y, splitter=0.2)
        >>> display = report.inspection.impurity_decrease()
        >>> display.frame()
              feature  importance
        0  Feature #0     0.10...
        1  Feature #1     0.32...
        2  Feature #2     0.08...
        3  Feature #3     0.48...
        4  Feature #4     0.00...
        """
        return ImpurityDecreaseDisplay._compute_data_for_display(
            estimator=self._parent.estimator_,
            name=self._parent.estimator_name_,
            report_type=self._parent._report_type,
        )

    def permutation_importance(
        self,
        *,
        data_source: DataSource = "test",
        at_step: int | str = 0,
        metric: Metric | list[Metric] | dict[str, Metric] | None = None,
        n_repeats: int = 5,
        max_samples: float = 1.0,
        n_jobs: int | None = None,
        seed: int | None = None,
    ) -> PermutationImportanceDisplay:
        """Display the permutation feature importance.

        This computes the permutation importance using sklearn's
        :func:`~sklearn.inspection.permutation_importance` function,
        which consists in permuting the values of one feature and comparing
        the value of `metric` between with and without the permutation, which gives an
        indication on the impact of the feature.

        By default, `seed` is set to `None`, which means the function will
        return a different result at every call. In that case, the results are not
        cached. If you wish to take advantage of skore's caching capabilities, make
        sure you set the `seed` parameter.

        Parameters
        ----------
        data_source : {"test", "train"}, default="test"
            The data source to use.

            - "test" : use the test set provided when creating the report.
            - "train" : use the train set provided when creating the report.

        at_step : int or str, default=0
            If the estimator is a :class:`~sklearn.pipeline.Pipeline`, at which step of
            the pipeline the importance is computed. If `n`, then the features that
            are evaluated are the ones *right before* the `n`-th step of the pipeline.
            For instance,

            - If 0, compute the importance just before the start of the pipeline (i.e.
              the importance of the raw input features).
            - If -1, compute the importance just before the end of the pipeline (i.e.
              the importance of the fully engineered features, just before the actual
              prediction step).

            If a string, will be searched among the pipeline's `named_steps`.

            Has no effect if the estimator is not a :class:`~sklearn.pipeline.Pipeline`.

        metric : str, callable, scorer, or list of such instances or dict of such \
            instances, default=None
            The metric to pass to :func:`~sklearn.inspection.permutation_importance`.
            The possible values (whether or not in a list) are:

            - if a string, either one of the built-in metrics or a scikit-learn scorer
              name. You can get the possible list of string using
              `report.metrics.help()` or :func:`sklearn.metrics.get_scorer_names` for
              the built-in metrics or the scikit-learn scorers, respectively.
            - if a callable, it should take as arguments `y_true`, `y_pred` as the two
              first arguments. Additional arguments can be passed as keyword arguments
              and will be forwarded with `metric_kwargs`. No favorability indicator can
              be displayed in this case.
            - if the callable API is too restrictive (e.g. need to pass
              same parameter name with different values), you can use scikit-learn
              scorers as provided by :func:`sklearn.metrics.make_scorer`. In this case,
              the metric favorability will only be displayed if it is given explicitly
              via `make_scorer`'s `greater_is_better` parameter.

        n_repeats : int, default=5
            Number of times to permute a feature.

        max_samples : int or float, default=1.0
            The number of samples to draw from `X` to compute feature importance
            in each repeat (without replacement).

            - If int, then draw max_samples samples.
            - If float, then draw max_samples * X.shape[0] samples.
            - If max_samples is equal to 1.0 or X.shape[0], all samples will be used.

            While using this option may provide less accurate importance estimates,
            it keeps the method tractable when evaluating feature importance on
            large datasets. In combination with n_repeats, this allows to control
            the computational speed vs statistical accuracy trade-off of this method.

        n_jobs : int or None, default=None
            Number of jobs to run in parallel. -1 means using all processors.

        seed : int or None, default=None
            The seed used to initialize the random number generator used for the
            permutations.

        Returns
        -------
        :class:`PermutationImportanceDisplay`
            The permutation importance display.

        Examples
        --------
        >>> from sklearn.datasets import make_regression
        >>> from sklearn.linear_model import Ridge
        >>> from skore import evaluate
        >>> X, y = make_regression(n_features=3, random_state=0)
        >>> regressor = Ridge()
        >>> report = evaluate(regressor, X, y, splitter=0.2)

        >>> report.inspection.permutation_importance(
        ...    n_repeats=2,
        ...    seed=0,
        ... ).frame(aggregate=None)
          data_source metric     feature  repetition     value
        0        test     r2  Feature #0           1  0.76...
        1        test     r2  Feature #1           1  1.39...
        2        test     r2  Feature #2           1  0.01...
        3        test     r2  Feature #0           2  0.80...
        4        test     r2  Feature #1           2  1.72...
        5        test     r2  Feature #2           2  0.02...

        >>> report.inspection.permutation_importance(
        ...    metric=["r2", "neg_mean_squared_error"],
        ...    n_repeats=2,
        ...    seed=0,
        ... ).frame(aggregate=None)
            data_source                  metric     feature  repetition      value
        0         test                      r2  Feature #0           1     0.76...
        1         test                      r2  Feature #1           1     1.39...
        2         test                      r2  Feature #2           1     0.01...
        3         test                      r2  Feature #0           2     0.80...
        4         test                      r2  Feature #1           2     1.72...
        5         test                      r2  Feature #2           2     0.02...
        6         test  neg_mean_squared_error  Feature #0           1  2561.20...
        7         test  neg_mean_squared_error  Feature #1           1  4686.07...
        8         test  neg_mean_squared_error  Feature #2           1    66.77...
        9         test  neg_mean_squared_error  Feature #0           2  2695.93...
        10        test  neg_mean_squared_error  Feature #1           2  5802.57...
        11        test  neg_mean_squared_error  Feature #2           2    93.41...

        >>> report.inspection.permutation_importance(
        ...    n_repeats=2,
        ...    seed=0,
        ... ).frame()
          data_source metric     feature  value_mean  value_std
        0        test     r2  Feature #0     0.78...    0.02...
        1        test     r2  Feature #1     1.56...    0.23...
        2        test     r2  Feature #2     0.02...    0.00...

        >>> # Compute the importance at the end of feature engineering pipeline
        >>> from sklearn.pipeline import make_pipeline
        >>> from sklearn.preprocessing import StandardScaler
        >>> pipeline = make_pipeline(StandardScaler(), Ridge())
        >>> pipeline_report = evaluate(pipeline, X, y, splitter=0.2)
        >>> pipeline_report.inspection.permutation_importance(
        ...    n_repeats=2,
        ...    seed=0,
        ...    at_step=-1,
        ... ).frame()
          data_source metric feature  value_mean  value_std
        0        test     r2      x0     0.78...    0.02...
        1        test     r2      x1     1.55...    0.23...
        2        test     r2      x2     0.02...    0.00...

        >>> pipeline_report.inspection.permutation_importance(
        ...    n_repeats=2,
        ...    seed=0,
        ...    at_step="ridge",
        ... ).frame()
          data_source metric feature  value_mean  value_std
        0        test     r2      x0     0.78...    0.02...
        1        test     r2      x1     1.55...    0.23...
        2        test     r2      x2     0.02...    0.00...

        Notes
        -----
        Even if pipeline components output sparse arrays, these will be made dense.
        """
        X_, y_true = self._get_X_y(data_source=data_source)

        # NOTE: to temporary improve the `project.put` UX, we always store the
        # permutation importance into the cache dictionary even when seed is None.
        # Be aware that if seed is None, we still trigger the computation for all cases.
        # We only store it such that when we serialize to send to the hub, we only
        # fetch for the cache store instead of recomputing it because it is expensive.
        # FIXME: the workaround above should be removed once we are able to trigger
        # computation on the server side of skore-hub.

        if seed is not None and not isinstance(seed, int):
            raise ValueError(f"seed must be an integer or None; got {type(seed)}")

        # n_jobs should not be in cache
        kwargs = {
            "at_step": at_step,
            "metric": metric,
            "n_repeats": n_repeats,
            "max_samples": max_samples,
            "seed": seed,
        }
        cache_key = make_cache_key(data_source, "permutation_importance", kwargs)

        # NOTE: avoid to fetch from the cache if the seed is None because we want
        # to trigger the computation in this case. We only have the permutation
        # stored as a workaround for the serialization for skore-hub as explained
        # earlier.
        display = None if seed is None else self._parent._cache.get(cache_key)
        if display is None:
            display = PermutationImportanceDisplay._compute_data_for_display(
                data_source=data_source,
                estimator=self._parent.estimator_,
                name=self._parent.estimator_name_,
                X=X_,
                y=y_true,
                at_step=at_step,
                metric=metric,
                n_repeats=n_repeats,
                max_samples=max_samples,
                n_jobs=n_jobs,
                seed=seed,
                report_type=self._parent._report_type,
            )

            if cache_key is not None:
                # NOTE: for the moment, we will always store the permutation importance
                self._parent._cache[cache_key] = display

        return display

    ####################################################################################
    # Methods related to the help tree
    ####################################################################################

    def __repr__(self) -> str:
        """Return a string representation using rich."""
        return self._rich_repr(class_name="skore.EstimatorReport.inspection")



from typing import Literal

from skrub import _dataframe as sbd

from skore._externals._pandas_accessors import DirNamesMixin
from skore._sklearn._base import _BaseAccessor
from skore._sklearn._cross_validation.report import CrossValidationReport
from skore._sklearn._plot import TableReportDisplay
from skore._utils._dataframe import _normalize_X_as_dataframe, _normalize_y_as_dataframe


class _DataAccessor(_BaseAccessor[CrossValidationReport], DirNamesMixin):
    """
    The data accessor helps you to get insights about the dataset used.

    It provides methods to create plots and to visualise the dataset.
    """

    def __init__(self, parent: CrossValidationReport) -> None:
        super().__init__(parent)

    def _retrieve_data_as_frame(
        self,
        with_y: bool,
    ):
        """Retrieve data as DataFrame.

        Parameters
        ----------
        with_y : bool
            Whether we should check that `y` is not None.

        Returns
        -------
        X : DataFrame
            The input data.

        y : DataFrame or None
            The target data.
        """
        X = self._parent.X
        y = self._parent.y

        X = _normalize_X_as_dataframe(X)

        if with_y:
            if y is None:
                raise ValueError("y is required when `with_y=True`.")

            y = _normalize_y_as_dataframe(y)

        return X, y

    def _prepare_dataframe_for_display(
        self,
        *,
        with_y: bool | None = None,
        subsample: int | None = None,
        subsample_strategy: Literal["head", "random"] = "head",
        seed: int | None = None,
    ):
        """Return features (and target when ``with_y``) as a single DataFrame.

        When ``with_y`` is ``None``, it defaults to supervised tasks only
        (``ml_task != "clustering"``): for clustering, only ``X`` is returned.
        """
        if subsample_strategy not in (subsample_strategy_options := ("head", "random")):
            raise ValueError(
                f"'subsample_strategy' options are {subsample_strategy_options!r}, got "
                f"{subsample_strategy}."
            )

        if with_y is None:
            with_y = self._parent.ml_task != "clustering"

        X, y = self._retrieve_data_as_frame(with_y)
        df = sbd.concat(X, y, axis=1) if with_y else X

        if subsample:
            if subsample_strategy == "head":
                df = sbd.head(df, subsample)
            else:  # subsample_strategy == "random":
                df = sbd.sample(df, subsample, seed=seed)

        return df

    def analyze(
        self,
        with_y: bool = True,
        subsample: int | None = None,
        subsample_strategy: Literal["head", "random"] = "head",
        seed: int | None = None,
    ) -> TableReportDisplay:
        """Plot dataset statistics.

        Parameters
        ----------
        with_y : bool, default=True
            Whether to include the target variable in the analysis. If True, the target
            variable is concatenated horizontally to the features.

        subsample : int, default=None
            The number of points to subsample the dataframe hold by the display, using
            the strategy set by ``subsample_strategy``. It must be a strictly positive
            integer. If ``None``, no subsampling is applied.

        subsample_strategy : {'head', 'random'}, default='head',
            The strategy used to subsample the dataframe hold by the display. It only
            has an effect when ``subsample`` is not None.

            - If ``'head'``: subsample by taking the ``subsample`` first points of the
              dataframe, similar to Pandas: ``df.head(n)``.
            - If ``'random'``: randomly subsample the dataframe by using a uniform
              distribution. The random seed is controlled by ``random_state``.

        seed : int, default=None
            The random seed to use when randomly subsampling. It only has an effect when
            ``subsample`` is not None and ``subsample_strategy='random'``.

        Returns
        -------
        :class:`TableReportDisplay`
            A display object containing the dataset statistics and plots.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import evaluate
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> classifier = LogisticRegression()
        >>> report = evaluate(classifier, X, y, splitter=2)
        >>> report.data.analyze().frame()
        """
        df = self._prepare_dataframe_for_display(
            with_y=with_y,
            subsample=subsample,
            subsample_strategy=subsample_strategy,
            seed=seed,
        )
        return TableReportDisplay._compute_data_for_display(df)

    ####################################################################################
    # Methods related to the help tree
    ####################################################################################

    def __repr__(self) -> str:
        """Return a string representation using rich."""
        return self._rich_repr(class_name="skore.CrossValidationReport.data")



from __future__ import annotations

import uuid
from collections import Counter
from collections.abc import Iterable
from typing import TYPE_CHECKING, Literal, cast

import joblib
import numpy as np
import pandas as pd
from numpy.typing import ArrayLike

from skore._externals._pandas_accessors import DirNamesMixin
from skore._sklearn._base import _BaseReport
from skore._sklearn._cross_validation.report import CrossValidationReport
from skore._sklearn._estimator.report import EstimatorReport
from skore._sklearn.types import PositiveLabel
from skore._utils._progress_bar import track
from skore._utils.repr.data import get_documentation_url
from skore._utils.repr.html_repr import render_template

if TYPE_CHECKING:
    from skore._sklearn._comparison.inspection_accessor import (
        _InspectionAccessor,
    )
    from skore._sklearn._comparison.metrics_accessor import _MetricsAccessor

    ComparisonReportType = Literal[
        "comparison-estimator",
        "comparison-cross-validation",
    ]


class ComparisonReport(_BaseReport, DirNamesMixin):
    """Report for comparing reports.

    This object can be used to compare several :class:`skore.EstimatorReport` instances,
    or several :class:`~skore.CrossValidationReport` instances.

    Refer to the :ref:`comparison_report` section of the user guide for more details.

    .. caution::
       Reports passed to `ComparisonReport` are not copied. If you pass
       a report to `ComparisonReport`, and then modify the report outside later, it
       will affect the report stored inside the `ComparisonReport` as well, which
       can lead to inconsistent results. For this reason, modifying reports after
       creation is strongly discouraged.

    Parameters
    ----------
    reports : list of reports or dict
        Reports to compare. If a dict, keys will be used to label the estimators;
        if a list, the labels are computed from the estimator class names.
        Expects at least two reports to compare, with the same test target.

    n_jobs : int, default=None
        Number of jobs to run in parallel. Training the estimators and computing
        the scores are parallelized.
        When accessing some methods of the `ComparisonReport`, the `n_jobs`
        parameter is used to parallelize the computation.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.

    Attributes
    ----------
    reports_ : dict mapping names to reports
        The compared reports.

    See Also
    --------
    skore.EstimatorReport
        Report for a fitted estimator.

    skore.CrossValidationReport
        Report for the cross-validation of an estimator.

    Examples
    --------
    >>> from sklearn.datasets import make_classification
    >>> from skore import train_test_split
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skore import ComparisonReport, EstimatorReport
    >>> X, y = make_classification(random_state=42)
    >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
    >>> estimator_1 = LogisticRegression()
    >>> estimator_report_1 = EstimatorReport(estimator_1, **split_data)
    >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
    >>> estimator_report_2 = EstimatorReport(estimator_2, **split_data)
    >>> report = ComparisonReport([estimator_report_1, estimator_report_2])
    >>> report.reports_
    {'LogisticRegression_1': ..., 'LogisticRegression_2': ...}
    >>> report = ComparisonReport(
    ...     {"model1": estimator_report_1, "model2": estimator_report_2}
    ... )
    >>> report.reports_
    {'model1': ..., 'model2': ...}

    >>> from sklearn.datasets import make_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skore import ComparisonReport, CrossValidationReport
    >>> X, y = make_classification(random_state=42)
    >>> estimator_1 = LogisticRegression()
    >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
    >>> report_1 = CrossValidationReport(estimator_1, X, y)
    >>> report_2 = CrossValidationReport(estimator_2, X, y)
    >>> report = ComparisonReport([report_1, report_2])
    >>> report = ComparisonReport({"model1": report_1, "model2": report_2})
    """

    _ACCESSOR_CONFIG: dict[str, dict[str, str]] = {
        "metrics": {"name": "metrics"},
        "inspection": {"name": "inspection"},
    }
    metrics: _MetricsAccessor
    inspection: _InspectionAccessor

    _report_type: ComparisonReportType

    @staticmethod
    def _validate_reports(
        reports: (
            list[EstimatorReport]
            | dict[str, EstimatorReport]
            | list[CrossValidationReport]
            | dict[str, CrossValidationReport]
        ),
    ) -> tuple[
        dict[str, EstimatorReport] | dict[str, CrossValidationReport],
        ComparisonReportType,
        PositiveLabel,
    ]:
        """Validate that reports are in the right format for comparison.

        Parameters
        ----------
        reports : list of reports or dict
            The reports to be validated.

        Returns
        -------
        dict
            The validated reports.
        {"comparison-estimator", "comparison-cross-validation"}
            The report type for the comparison.
        int, float, bool, str or None
            The positive label used in the different reports.
        """
        if not isinstance(reports, Iterable):
            raise TypeError(
                f"Expected reports to be a list or dict; got {type(reports)}"
            )

        if len(reports) < 2:
            raise ValueError(
                f"Expected at least 2 reports to compare; got {len(reports)}"
            )

        if isinstance(reports, list):
            report_names = None
            reports_list = reports
        else:  # dict
            report_names = list(reports.keys())
            for key in report_names:
                if not isinstance(key, str):
                    raise TypeError(
                        f"Expected all report names to be strings; got {type(key)}"
                    )
            reports_list = cast(
                list[EstimatorReport] | list[CrossValidationReport],
                list(reports.values()),
            )

        report_type: ComparisonReportType
        if all(isinstance(report, EstimatorReport) for report in reports_list):
            reports_list = cast(list[EstimatorReport], reports_list)
            report_type = "comparison-estimator"

            test_dataset_hashes = {
                joblib.hash(report.y_test)
                for report in reports_list
                if report.y_test is not None
            }
            if len(test_dataset_hashes) > 1:
                raise ValueError(
                    "Expected all estimators to share the same test targets."
                )

        elif all(isinstance(report, CrossValidationReport) for report in reports_list):
            reports_list = cast(list[CrossValidationReport], reports_list)
            report_type = "comparison-cross-validation"
        else:
            raise TypeError(
                f"Expected list or dict of {EstimatorReport.__name__} "
                f"or list of dict of {CrossValidationReport.__name__}"
            )

        if len({id(report) for report in reports_list}) < len(reports_list):
            raise ValueError("Expected reports to be distinct objects")

        ml_tasks = {report._ml_task for report in reports_list}
        if len(ml_tasks) > 1:
            raise ValueError(
                f"Expected all estimators to have the same ML usecase; got {ml_tasks}"
            )

        if ml_tasks == {"binary-classification"}:
            pos_labels = {report.pos_label for report in reports_list}
            if len(pos_labels) > 1:
                raise ValueError(
                    "Expected all estimators to have the same positive label. "
                    f"Got {pos_labels}."
                )
            pos_label = pos_labels.pop()
        else:
            pos_label = None

        if report_names is None:
            deduped_report_names = _deduplicate_report_names(
                [report.estimator_name_ for report in reports_list]
            )
        else:
            deduped_report_names = report_names

        reports_dict = cast(
            dict[str, EstimatorReport] | dict[str, CrossValidationReport],
            dict(zip(deduped_report_names, reports_list, strict=True)),
        )

        return reports_dict, report_type, pos_label

    def __init__(
        self,
        reports: (
            list[EstimatorReport]
            | dict[str, EstimatorReport]
            | list[CrossValidationReport]
            | dict[str, CrossValidationReport]
        ),
        *,
        n_jobs: int | None = None,
    ) -> None:
        """
        ComparisonReport instance initializer.

        Notes
        -----
        We check that the estimator reports can be compared:
        - all reports are estimator reports,
        - all estimators are in the same ML use case,
        - all estimators have non-empty X_test and y_test,
        - all estimators have the same X_test and y_test.
        """
        super().__init__()
        self.reports_, self._report_type, self._pos_label = (
            ComparisonReport._validate_reports(reports)
        )

        self.n_jobs = n_jobs
        self._ml_task = next(iter(self.reports_.values()))._ml_task  # type: ignore

    def clear_cache(self) -> None:
        """Clear the cache.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import ComparisonReport, EstimatorReport
        >>> X, y = make_classification(random_state=42)
        >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
        >>> estimator_1 = LogisticRegression()
        >>> estimator_report_1 = EstimatorReport(estimator_1, **split_data)
        >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
        >>> estimator_report_2 = EstimatorReport(estimator_2, **split_data)
        >>> report = ComparisonReport([estimator_report_1, estimator_report_2])
        >>> report.cache_predictions()
        >>> report.clear_cache()
        """
        for report in self.reports_.values():
            report.clear_cache()

    def cache_predictions(
        self,
        response_methods: Literal[
            "auto", "predict", "predict_proba", "decision_function"
        ] = "auto",
        n_jobs: int | None = None,
    ) -> None:
        """Cache the predictions for sub-estimators reports.

        Parameters
        ----------
        response_methods : {"auto", "predict", "predict_proba", "decision_function"},\
                default="auto"
            The methods to use to compute the predictions.

        n_jobs : int, default=None
            The number of jobs to run in parallel. If `None`, we use the `n_jobs`
            parameter when initializing the report.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import ComparisonReport, EstimatorReport
        >>> X, y = make_classification(random_state=42)
        >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
        >>> estimator_1 = LogisticRegression()
        >>> estimator_report_1 = EstimatorReport(estimator_1, **split_data)
        >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
        >>> estimator_report_2 = EstimatorReport(estimator_2, **split_data)
        >>> report = ComparisonReport([estimator_report_1, estimator_report_2])
        >>> report.cache_predictions()
        """
        if n_jobs is None:
            n_jobs = self.n_jobs

        for report in track(
            self.reports_.values(),
            description="Estimator predictions",
            total=len(self.reports_),
        ):
            report.cache_predictions(response_methods=response_methods, n_jobs=n_jobs)

    def get_predictions(
        self,
        *,
        data_source: Literal["train", "test"],
        response_method: Literal[
            "predict", "predict_proba", "decision_function"
        ] = "predict",
    ) -> list[ArrayLike] | list[list[ArrayLike]]:
        """Get predictions from the underlying reports.

        This method has the advantage to reload from the cache if the predictions
        were already computed in a previous call.

        Parameters
        ----------
        data_source : {"test", "train"}, default="test"
            The data source to use.

            - "test" : use the test set provided when creating the report.
            - "train" : use the train set provided when creating the report.

        response_method : {"predict", "predict_proba", "decision_function"}, \
                default="predict"
            The response method to use to get the predictions.

        Returns
        -------
        list of np.ndarray of shape (n_samples,) or (n_samples, n_classes) or list of \
                such lists
            The predictions for each :class:`~skore.EstimatorReport` or
            :class:`~skore.CrossValidationReport`.

        Raises
        ------
        ValueError
            If the data source is invalid.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from skore import train_test_split
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import ComparisonReport, EstimatorReport
        >>> X, y = make_classification(random_state=42)
        >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
        >>> estimator_1 = LogisticRegression()
        >>> estimator_report_1 = EstimatorReport(estimator_1, **split_data)
        >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
        >>> estimator_report_2 = EstimatorReport(estimator_2, **split_data)
        >>> report = ComparisonReport([estimator_report_1, estimator_report_2])
        >>> report.cache_predictions()
        >>> predictions = report.get_predictions(data_source="test")
        >>> print([split_predictions.shape for split_predictions in predictions])
        [(25,), (25,)]
        """
        return [  # type: ignore
            report.get_predictions(
                data_source=data_source,
                response_method=response_method,
            )
            for report in self.reports_.values()
        ]

    def create_estimator_report(
        self,
        *,
        report_key: str,
        X_test: ArrayLike,
        y_test: ArrayLike,
    ) -> EstimatorReport:
        """Create an estimator report from one of the reports in the comparison.

        This method creates a new :class:`~skore.EstimatorReport` with the same
        estimator and the same data as the chosen report. It is useful to evaluate and
        deploy a model that was deemed optimal during the comparison. Provide a held out
        test set to properly evaluate the performance of the model.

        Parameters
        ----------
        report_key : str
            The key associated with the estimator to create a report for, as stored in
            the :attr:`reports_` attribute of the :class:`~skore.ComparisonReport`.
            List the available keys with `reports_.keys()`.

        X_test : {array-like, sparse matrix} of shape (n_samples, n_features)
            Testing data. It should have the same structure as the training data.

        y_test : array-like of shape (n_samples,) or (n_samples, n_outputs) or None
            Testing target.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import ComparisonReport, CrossValidationReport
        >>> X, y = make_classification(random_state=42)
        >>> X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=42)
        >>> linear_report = CrossValidationReport(
        ...     LogisticRegression(random_state=42), X_train, y_train
        ... )
        >>> forest_report = CrossValidationReport(
        ...     RandomForestClassifier(random_state=42), X_train, y_train
        ... )
        >>> comparison_report = ComparisonReport([linear_report, forest_report])
        >>> summary = comparison_report.metrics.summarize().frame()

        >>> # Notice that e.g. the RandomForestClassifier performs best
        >>> final_report = comparison_report.create_estimator_report(
        ...     report_key="RandomForestClassifier", X_test=X_test, y_test=y_test
        ... )
        >>> final_report.metrics.summarize().frame()

        Returns
        -------
        report : :class:`~skore.EstimatorReport`
            The estimator report.
        """
        if report_key not in self.reports_:
            raise ValueError(
                f"Estimator with key {report_key} not found in the comparison report. "
                f"Available keys are: {list(self.reports_.keys())}."
            )

        if isinstance(self.reports_[report_key], CrossValidationReport):
            return cast(
                CrossValidationReport, self.reports_[report_key]
            ).create_estimator_report(X_test=X_test, y_test=y_test)

        estimator_report = cast(EstimatorReport, self.reports_[report_key])
        X_concat = (
            pd.concat([estimator_report._X_train, estimator_report._X_test])
            if isinstance(estimator_report._X_train, pd.DataFrame)
            else np.concatenate([estimator_report._X_train, estimator_report._X_test])
        )
        y_concat = (
            pd.concat([estimator_report._y_train, estimator_report._y_test])
            if isinstance(estimator_report._y_train, (pd.DataFrame, pd.Series))
            else np.concatenate([estimator_report._y_train, estimator_report._y_test])
        )
        report = EstimatorReport(
            estimator_report.estimator,
            X_train=X_concat,
            y_train=y_concat,
            X_test=X_test,
            y_test=y_test,
            pos_label=self._pos_label,
        )
        return report

    @property
    def pos_label(self) -> PositiveLabel | None:
        return self._pos_label

    ####################################################################################
    # Methods related to the help and repr
    ####################################################################################

    def _get_help_title(self) -> str:
        return "Tools to compare estimators"

    def _get_help_legend(self) -> str:
        return (
            "[cyan](↗︎)[/cyan] higher is better [orange1](↘︎)[/orange1] lower is better"
        )

    def __repr__(self) -> str:
        """Return a string representation."""
        return f"{self.__class__.__name__}(...)"

    def _repr_html_(self) -> str:
        """HTML representation with a selector to inspect one compared report."""
        metrics_html = self.metrics.summarize(data_source="test").frame()._repr_html_()

        comparison_reports = []
        for label, report in self.reports_.items():
            fragments = report._html_repr_fragments()
            comparison_reports.append(
                {
                    "label": label,
                    "estimator_display": fragments["estimator_display"],
                    "table_report": fragments["table_report"],
                }
            )

        container_id = f"skore-comparison-report-{uuid.uuid4().hex[:8]}"
        help_doc_url = get_documentation_url(obj=self, method_name="help")
        report_class_name = self.__class__.__name__
        metrics_accessor_doc_url = get_documentation_url(
            obj=self, accessor_name="metrics"
        )
        inspection_accessor_doc_url = get_documentation_url(
            obj=self, accessor_name="inspection"
        )
        return render_template(
            "comparison_report.html.j2",
            {
                "container_id": container_id,
                "metrics_summary": metrics_html,
                "comparison_reports": comparison_reports,
                "help_doc_url": help_doc_url,
                "report_class_name": report_class_name,
                "metrics_accessor_doc_url": metrics_accessor_doc_url,
                "inspection_accessor_doc_url": inspection_accessor_doc_url,
            },
        )

    def _repr_mimebundle_(self, **kwargs):
        """Mime bundle used by Jupyter kernels to display the report."""
        output = {"text/plain": repr(self)}
        output["text/html"] = self._repr_html_()
        return output


def _deduplicate_report_names(report_names: list[str]) -> list[str]:
    """De-duplicate report names that appear several times.

    Leave the other report names alone.

    Parameters
    ----------
    report_names : list of str
        The list of report names to be checked.

    Returns
    -------
    list of str
        The de-duplicated list of report names.

    Examples
    --------
    >>> _deduplicate_report_names(['a', 'b'])
    ['a', 'b']
    >>> _deduplicate_report_names(['a', 'a'])
    ['a_1', 'a_2']
    >>> _deduplicate_report_names(['a', 'b', 'a'])
    ['a_1', 'b', 'a_2']
    >>> _deduplicate_report_names(['a', 'b', 'a', 'b'])
    ['a_1', 'b_1', 'a_2', 'b_2']
    >>> _deduplicate_report_names([])
    []
    >>> _deduplicate_report_names(['a'])
    ['a']
    """
    counts = Counter(report_names)
    if len(report_names) == len(counts):
        return report_names

    names = report_names.copy()
    seen: Counter = Counter()
    for i in range(len(names)):
        name = names[i]
        seen[name] += 1
        if counts[name] > 1:
            names[i] = f"{name}_{seen[name]}"
    return names



from __future__ import annotations

from typing import Any, Literal

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.figure import Figure
from sklearn.base import BaseEstimator, is_classifier
from sklearn.compose import TransformedTargetRegressor
from sklearn.pipeline import Pipeline

from skore._sklearn._plot.base import BOXPLOT_STYLE, DisplayMixin
from skore._sklearn._plot.inspection.utils import (
    _decorate_matplotlib_axis,
    select_k_features,
    sort_features,
)
from skore._sklearn.feature_names import _get_feature_names
from skore._sklearn.types import Aggregate, ReportType
from skore._utils._index import flatten_multi_index


class CoefficientsDisplay(DisplayMixin):
    """Display to inspect the coefficients of linear models.

    Parameters
    ----------
    coefficients : DataFrame | list[DataFrame]
        The coefficients data to display. The columns are:

        - `estimator`
        - `split`
        - `feature`
        - `label` or `output` (classification vs. regression)
        - `coefficient`

    report_type : {"estimator", "cross-validation", "comparison-estimator", \
            "comparison-cross-validation"}
        Report type from which the display is created.

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skore import evaluate
    >>> iris = load_iris(as_frame=True)
    >>> X, y = iris.data, iris.target
    >>> y = iris.target_names[y]
    >>> report = evaluate(LogisticRegression(), X, y, splitter=0.2)
    >>> display = report.inspection.coefficients()
    >>> display.frame()
                  feature       label  coefficient
    0           Intercept      setosa        9.2...
    1   sepal length (cm)      setosa       -0.3...
    2    sepal width (cm)      setosa        0.8...
    3   petal length (cm)      setosa       -2.3...
    4    petal width (cm)      setosa       -1.0...
    5           Intercept  versicolor        2.3...
    6   sepal length (cm)  versicolor        0.4...
    7    sepal width (cm)  versicolor       -0.3...
    8   petal length (cm)  versicolor       -0.1...
    9    petal width (cm)  versicolor       -0.7...
    10          Intercept   virginica      -11.5...
    11  sepal length (cm)   virginica       -0.0...
    12   sepal width (cm)   virginica       -0.5...
    13  petal length (cm)   virginica        2.5...
    14   petal width (cm)   virginica        1.7...
    """

    _default_barplot_kwargs: dict[str, Any] = {
        "aspect": 2,
        "height": 6,
        "palette": "tab10",
    }
    _default_stripplot_kwargs: dict[str, Any] = {
        "alpha": 0.5,
        "aspect": 2,
        "height": 6,
        "palette": "tab10",
    }
    _default_boxplot_kwargs: dict[str, Any] = {
        "whis": 1e10,
        **BOXPLOT_STYLE,
    }

    def __init__(self, *, coefficients: pd.DataFrame, report_type: ReportType):
        self.coefficients = coefficients
        self.report_type = report_type

    def frame(
        self,
        *,
        aggregate: Aggregate | None = ("mean", "std"),
        include_intercept: bool = True,
        select_k: int | None = None,
        sorting_order: Literal["descending", "ascending", None] = None,
    ):
        """Get the coefficients in a dataframe format.

        The returned dataframe is not going to contain constant columns or columns
        containing only NaN values.

        Parameters
        ----------
        include_intercept : bool, default=True
            Whether or not to include the intercept in the dataframe.

        select_k : int, default=None
            Select features based on absolute coefficient values:

            - Positive values: select the `select_k` features with largest absolute
              coefficients
            - Negative values: select the `-select_k` features with smallest absolute
              coefficients

            Selection is performed independently within each group:

            - Single estimator reports: For binary classification or single-output
              regression, selection is global. For multiclass classification or
              multi-output regression, selection is performed independently per
              class/output.
            - Cross-validation reports: Grouping follows the same rules as single
              estimator reports. Within each group, features are ranked by the mean
              absolute coefficient values across folds.
            - Comparison reports: Selection is performed independently per estimator,
              and per class/output if applicable.

        sorting_order : {"descending", "ascending", None}, default=None
            Sort features by absolute coefficient values:

            - "descending": largest absolute values first
            - "ascending": smallest absolute values first
            - None: preserve original order

            Can be used independently of `select_k`. Sorting is performed within the
            same groups as selection.

        Returns
        -------
        DataFrame
            Dataframe containing the coefficients of the linear model.

        Examples
        --------
        >>> from sklearn.datasets import load_iris
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import evaluate
        >>> iris = load_iris(as_frame=True)
        >>> X, y = iris.data, iris.target
        >>> y = iris.target_names[y]
        >>> report = evaluate(LogisticRegression(), X, y, splitter=0.2)
        >>> display = report.inspection.coefficients()
        >>> display.frame()
                      feature       label  coefficient
        0           Intercept      setosa        9.2...
        1   sepal length (cm)      setosa       -0.3...
        2    sepal width (cm)      setosa        0.8...
        3   petal length (cm)      setosa       -2.3...
        4    petal width (cm)      setosa       -1.0...
        5           Intercept  versicolor        2.3...
        6   sepal length (cm)  versicolor        0.4...
        7    sepal width (cm)  versicolor       -0.3...
        8   petal length (cm)  versicolor       -0.1...
        9    petal width (cm)  versicolor       -0.7...
        10          Intercept   virginica      -11.5...
        11  sepal length (cm)   virginica       -0.0...
        12   sepal width (cm)   virginica       -0.5...
        13  petal length (cm)   virginica        2.5...
        14   petal width (cm)   virginica        1.7...
        """
        if self.report_type == "estimator":
            columns_to_drop = ["estimator", "split"]
        elif self.report_type == "cross-validation":
            columns_to_drop = ["estimator"]
        elif self.report_type == "comparison-estimator":
            columns_to_drop = ["split"]
        else:  # comparison-cross-validation
            columns_to_drop = []

        if self.coefficients["label"].isna().all():
            # regression problem
            columns_to_drop.append("label")
        if self.coefficients["output"].isna().all():
            # classification problem
            columns_to_drop.append("output")

        coefficients = self.coefficients.drop(columns=columns_to_drop)
        if not include_intercept:
            coefficients = coefficients.query("feature != 'Intercept'")

        if sorting_order is not None:
            coefficients = sort_features(
                coefficients,
                sorting_order,
                group_columns=self._get_columns_to_groupby(frame=coefficients),
                importance_column="coefficient",
                use_absolute=True,
            )
        if select_k is not None:
            coefficients = select_k_features(
                coefficients,
                select_k,
                group_columns=self._get_columns_to_groupby(frame=coefficients),
                importance_column="coefficient",
                use_absolute=True,
            )

        if aggregate is not None and "split" in coefficients.columns:
            group_by = [
                c
                for c in ["estimator", "feature", "label", "output"]
                if c in coefficients.columns
            ]
            coefficients = (
                coefficients.drop(columns=["split"])
                .groupby(group_by, sort=False, dropna=False)
                .aggregate(aggregate)
            ).reset_index()
            if isinstance(coefficients.columns, pd.MultiIndex):
                coefficients.columns = flatten_multi_index(coefficients.columns)

        return coefficients

    @DisplayMixin.style_plot
    def plot(
        self,
        *,
        include_intercept: bool = True,
        subplot_by: Literal["auto", "estimator", "label", "output"] | None = "auto",
        select_k: int | None = None,
        sorting_order: Literal["descending", "ascending", None] = None,
    ) -> Figure:
        """Plot the coefficients for the different features.

        Parameters
        ----------
        include_intercept : bool, default=True
            Whether or not to include the intercept in the dataframe.

        subplot_by : {"auto", "estimator", "label", "output"} or None, default="auto"
            The column to use for subplotting and dividing the coefficients into
            subplots. If "auto", not subplotting is performed apart from:

            - when comparing estimators in a multiclass classification or multi-output
              regression problem;
            - when comparing estimators for which the input features are different.

        select_k : int, default=None
            Select features based on absolute coefficient values:

            - Positive values: select the `select_k` features with largest absolute
              coefficients
            - Negative values: select the `-select_k` features with smallest absolute
              coefficients

            Selection is performed independently within each group as described in
            the `frame` method.

        sorting_order : {"descending", "ascending", None}, default=None
            Sort features by absolute coefficient values:

            - "descending": largest absolute values first
            - "ascending": smallest absolute values first
            - None: preserve original order

            Can be used independently of `select_k`. Sorting is performed within the
            same groups as selection.

        Returns
        -------
        matplotlib.figure.Figure
            Figure containing the coefficients plot.

        Examples
        --------
        >>> from sklearn.datasets import load_iris
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import evaluate
        >>> iris = load_iris(as_frame=True)
        >>> X, y = iris.data, iris.target
        >>> y = iris.target_names[y]
        >>> report = evaluate(LogisticRegression(), X, y, splitter=0.2)
        >>> display = report.inspection.coefficients()
        >>> display.plot()
        """
        return self._plot(
            include_intercept=include_intercept,
            subplot_by=subplot_by,
            select_k=select_k,
            sorting_order=sorting_order,
        )

    def _plot_matplotlib(
        self,
        *,
        include_intercept: bool = True,
        subplot_by: Literal["estimator", "label", "output"] | None = None,
        select_k: int | None = None,
        sorting_order: Literal["descending", "ascending", None] = None,
    ) -> Figure:
        """Dispatch the plotting function for matplotlib backend."""
        if select_k == 0:
            raise ValueError(
                "select_k=0 would produce an empty plot. Use a non-zero value or "
                "omit select_k to plot all features."
            )
        frame = self.frame(
            aggregate=None,
            include_intercept=include_intercept,
            select_k=select_k,
            sorting_order=sorting_order,
        )

        # Make copy of the dictionary since we are going to pop some keys later
        barplot_kwargs = self._default_barplot_kwargs.copy()
        boxplot_kwargs = self._default_boxplot_kwargs.copy()
        stripplot_kwargs = self._default_stripplot_kwargs.copy()

        if "comparison" in self.report_type:
            return self._plot_comparison(
                frame=frame,
                report_type=self.report_type,
                subplot_by=subplot_by,
                barplot_kwargs=barplot_kwargs,
                boxplot_kwargs=boxplot_kwargs,
                stripplot_kwargs=stripplot_kwargs,
            )
        # EstimatorReport or CrossValidationReport
        return self._plot_single_estimator(
            frame=frame,
            estimator_name=self.coefficients["estimator"][0],
            report_type=self.report_type,
            subplot_by=subplot_by,
            barplot_kwargs=barplot_kwargs,
            boxplot_kwargs=boxplot_kwargs,
            stripplot_kwargs=stripplot_kwargs,
        )

    @staticmethod
    def _get_columns_to_groupby(*, frame: pd.DataFrame) -> list[str]:
        """Get the available columns from which to group by."""
        columns_to_groupby = list[str]()
        if "estimator" in frame.columns:
            columns_to_groupby.append("estimator")
        if "label" in frame.columns:
            columns_to_groupby.append("label")
        if "output" in frame.columns:
            columns_to_groupby.append("output")
        return columns_to_groupby

    def _categorical_plot(
        self,
        *,
        frame: pd.DataFrame,
        report_type: ReportType,
        hue: str | None = None,
        col: str | None = None,
        barplot_kwargs: dict[str, Any] | None = None,
        boxplot_kwargs: dict[str, Any] | None = None,
        stripplot_kwargs: dict[str, Any] | None = None,
    ) -> Figure:
        if "estimator" in report_type:
            facet = sns.catplot(
                data=frame,
                x="coefficient",
                y="feature",
                hue=hue,
                col=col,
                kind="bar",
                **barplot_kwargs,
            )
        else:  # "cross-validation" in report_type
            facet = sns.catplot(
                data=frame,
                x="coefficient",
                y="feature",
                hue=hue,
                col=col,
                kind="strip",
                dodge=True,
                **stripplot_kwargs,
            ).map_dataframe(
                sns.boxplot,
                x="coefficient",
                y="feature",
                hue=hue,
                palette="tab10" if hue is not None else None,
                dodge=True,
                **boxplot_kwargs,
            )
        add_background_features = hue is not None

        figure = facet.figure
        ax_grid = facet.axes.squeeze()
        n_features = (
            [frame["feature"].nunique()]
            if col is None
            else [
                frame.query(f"{col} == '{col_value}'")["feature"].nunique()
                for col_value in frame[col].unique()
            ]
        )
        for ax, n_feature in zip(ax_grid.flatten(), n_features, strict=True):
            _decorate_matplotlib_axis(
                ax=ax,
                add_background_features=add_background_features,
                n_features=n_feature,
                xlabel="Magnitude of coefficient",
                ylabel="",
            )
        return figure

    def _plot_single_estimator(
        self,
        *,
        frame: pd.DataFrame,
        estimator_name: str,
        report_type: ReportType,
        subplot_by: Literal["auto", "estimator", "label", "output"] | None,
        barplot_kwargs: dict[str, Any],
        boxplot_kwargs: dict[str, Any],
        stripplot_kwargs: dict[str, Any],
    ) -> Figure:
        """Plot the coefficients for an `EstimatorReport` or a `CrossValidationReport`.

        An `EstimatorReport` will use a bar plot while a `CrossValidationReport` will
        use a box plot.

        Parameters
        ----------
        frame : pd.DataFrame
            The frame to plot.

        estimator_name : str
            The name of the estimator to plot.

        report_type : {"estimator", "cross-validation"}
            The type of report to plot.

        subplot_by : {"auto", "estimator", "label", "output"} or None
            The column to use for subplotting and dividing the coefficients into
            subplots.

        barplot_kwargs : dict
            Keyword arguments to be passed to :func:`seaborn.barplot` for
            rendering the coefficients with an :class:`~skore.EstimatorReport` or
            :class:`~skore.ComparisonReport` of :class:`~skore.EstimatorReport`.

        boxplot_kwargs : dict
            Keyword arguments to be passed to :func:`seaborn.boxplot` for
            rendering the coefficients with a :class:`~skore.CrossValidationReport` or
            :class:`~skore.ComparisonReport` of :class:`~skore.CrossValidationReport`.

        stripplot_kwargs : dict
            Keyword arguments to be passed to :func:`seaborn.stripplot` for
            rendering the coefficients with a :class:`~skore.CrossValidationReport` or
            :class:`~skore.ComparisonReport` of :class:`~skore.CrossValidationReport`.
        """
        # {"label"} or {"output"} or {}
        columns_to_groupby = self._get_columns_to_groupby(frame=frame)
        if subplot_by == "auto":
            subplot_by = None

        if subplot_by is not None and not len(columns_to_groupby):
            raise ValueError(
                "No columns to group by. `subplot_by` is expected to be None or 'auto'."
            )
        elif subplot_by is not None and subplot_by not in columns_to_groupby:
            raise ValueError(
                f"Column {subplot_by} not found in the frame. It should be one "
                f"of {', '.join(columns_to_groupby + ['auto', 'None'])}."
            )

        if subplot_by is None:
            hue = None if not len(columns_to_groupby) else columns_to_groupby[0]
            if hue is None:
                barplot_kwargs.pop("palette", None)
                stripplot_kwargs.pop("palette", None)
            col = None
        else:
            hue, col = None, subplot_by
            # we don't need the palette and we are at risk of raising an error or
            # deprecation warning if passing palette without a hue
            barplot_kwargs.pop("palette", None)
            stripplot_kwargs.pop("palette", None)

        figure = self._categorical_plot(
            frame=frame,
            report_type=report_type,
            hue=hue,
            col=col,
            barplot_kwargs=barplot_kwargs,
            boxplot_kwargs=boxplot_kwargs,
            stripplot_kwargs=stripplot_kwargs,
        )

        title = f"Coefficients of {estimator_name}"
        if subplot_by is not None:
            title += f" by {subplot_by}"
        figure.suptitle(title)
        return figure

    @staticmethod
    def _has_same_features(*, frame: pd.DataFrame) -> bool:
        """Check if the features are the same across all estimators."""
        grouped = {
            name: group["feature"].sort_values().tolist()
            for name, group in frame.groupby("estimator", sort=False)
        }
        _, reference_features = grouped.popitem()
        for group_features in grouped.values():
            if group_features != reference_features:
                return False
        return True

    def _plot_comparison(
        self,
        *,
        frame: pd.DataFrame,
        report_type: ReportType,
        subplot_by: Literal["auto", "estimator", "label", "output"] | None,
        barplot_kwargs: dict[str, Any],
        boxplot_kwargs: dict[str, Any],
        stripplot_kwargs: dict[str, Any],
    ) -> Figure:
        """Plot the coefficients for a `ComparisonReport`.

        Parameters
        ----------
        frame : pd.DataFrame
            The frame to plot.

        report_type : {"comparison-estimator", "comparison-cross-validation"}
            The type of report to plot.

        subplot_by : {"auto", "estimator", "label", "output"} or None
            The column to use for subplotting and dividing the coefficients into
            subplots. If `None`, an automatic choice is made depending on the type of
            reports at hand.

        barplot_kwargs : dict
            Keyword arguments to be passed to :func:`seaborn.barplot` for
            rendering the coefficients with an :class:`~skore.ComparisonReport` of
            :class:`~skore.EstimatorReport`.

        boxplot_kwargs : dict
            Keyword arguments to be passed to :func:`seaborn.boxplot` for
            rendering the coefficients with a :class:`~skore.ComparisonReport` of
            :class:`~skore.CrossValidationReport`.

        stripplot_kwargs : dict
            Keyword arguments to be passed to :func:`seaborn.stripplot` for
            rendering the coefficients with a :class:`~skore.ComparisonReport` of
            :class:`~skore.CrossValidationReport`.
        """
        # help mypy to understand the following variable types
        hue: str | None = None

        # {"estimator", "label"} or {"estimator", "output"} or {"estimator"}
        columns_to_groupby = self._get_columns_to_groupby(frame=frame)
        if subplot_by not in ("auto", None) and subplot_by not in columns_to_groupby:
            additional_subplot_by = ["auto"]
            if "label" not in frame.columns and "output" not in frame.columns:
                additional_subplot_by.append("None")

            raise ValueError(
                f"Column {subplot_by} not found in the frame. It should be one "
                f"of {', '.join(columns_to_groupby + additional_subplot_by)}."
            )
        elif subplot_by is None:
            if "label" in frame.columns:
                n_unique = frame["label"].nunique()
            elif "output" in frame.columns:
                n_unique = frame["output"].nunique()
            else:
                # binary classification or single output regression
                n_unique = 1
            if n_unique > 1:
                raise ValueError(
                    "There are multiple labels or outputs and `subplot_by` is `None`. "
                    "There is too much information to display on a single plot. "
                    "Please provide a column to group by using `subplot_by`."
                )

        has_same_features = self._has_same_features(frame=frame)
        if (frame.columns.isin(["label", "output"]).any() and subplot_by == "auto") or (
            not has_same_features and subplot_by == "auto"
        ):
            # default fallback on subplots by estimator
            # case 1: multiclass classification or multi-output regression
            # therefore, too many information to display on a single plot, by default
            # group by estimator
            # case 2: features cannot be compared across estimators and we therefore
            # need to subplots by estimator
            subplot_by = "estimator"
        elif subplot_by == "auto":
            subplot_by = None

        if subplot_by is None:
            hue, col = columns_to_groupby[0], None
            if not has_same_features and hue == "estimator":
                raise ValueError(
                    "The estimators have different features and should be plotted on "
                    "different axis using `subplot_by='estimator'`."
                )
        else:
            # infer if we should group by another column using hue
            hue_groupby = [col for col in columns_to_groupby if col != subplot_by]
            hue = hue_groupby[0] if len(hue_groupby) else None
            col = subplot_by

            if hue is None:
                barplot_kwargs.pop("palette", None)
                stripplot_kwargs.pop("palette", None)

            if not has_same_features and hue == "estimator":
                raise ValueError(
                    "The estimators have different features and should be plotted on "
                    "different axis using `subplot_by='estimator'`."
                )

        figure = self._categorical_plot(
            frame=frame,
            report_type=report_type,
            hue=hue,
            col=col,
            barplot_kwargs={"sharey": has_same_features} | barplot_kwargs,
            boxplot_kwargs=boxplot_kwargs,
            stripplot_kwargs={"sharey": has_same_features} | stripplot_kwargs,
        )

        title = "Coefficients"
        if subplot_by is not None:
            title += f" by {subplot_by}"
        figure.suptitle(title)
        return figure

    @classmethod
    def _compute_data_for_display(
        cls,
        *,
        estimator: BaseEstimator,
        name: str,
        report_type: ReportType,
    ) -> CoefficientsDisplay:
        """Compute the data for the display from a single estimator.

        Parameters
        ----------
        estimator : estimator
            The estimator to compute the data for.

        name : str
            The name of the estimator.

        report_type : {"estimator", "cross-validation", "comparison-estimator", \
                "comparison-cross-validation"}
            The type of report to compute the data for.

        Returns
        -------
        CoefficientsDisplay
            The data for the display.
        """
        if isinstance(estimator, Pipeline):
            preprocessor, predictor = estimator[:-1], estimator[-1]
        else:
            preprocessor, predictor = None, estimator

        if isinstance(predictor, TransformedTargetRegressor):
            predictor = predictor.regressor_

        coef = np.atleast_2d(predictor.coef_).T
        intercept = np.atleast_2d(predictor.intercept_)
        if coef.shape[1] != intercept.shape[1]:
            # it happens that with `fit_intercept=False` and a multi-output
            # regression problem, intercept is a single float. Thus, we need to
            # repeat it for each output
            intercept = np.repeat(intercept, coef.shape[1], axis=1)

        coef_data = np.concatenate([intercept, coef])

        feature_names = ["Intercept"] + _get_feature_names(
            predictor, transformer=preprocessor, n_features=coef.shape[0]
        )
        n_features = len(feature_names)

        index = pd.DataFrame(
            {
                "estimator": [name] * n_features,
                "split": [np.nan] * n_features,
                "feature": feature_names,
            }
        )

        if coef.shape[1] == 1:
            # binary or single output regression
            columns, require_melting = ["coefficient"], False
            index["label"], index["output"] = np.nan, np.nan
        else:
            require_melting = True
            if is_classifier(predictor):
                # multi-class classification
                columns, var_name = predictor.classes_.tolist(), "label"
                index["output"] = np.nan
            else:
                # multi-output regression
                columns, var_name = [f"{i}" for i in range(coef.shape[1])], "output"
                index["label"] = np.nan
            id_vars, value_name = index.columns.tolist(), "coefficient"

        coefficients = pd.concat(
            [index, pd.DataFrame(coef_data, columns=columns)], axis=1
        )
        if require_melting:
            coefficients = coefficients.melt(
                id_vars=id_vars,
                value_vars=columns,
                var_name=var_name,
                value_name=value_name,
            )

        return cls(coefficients=coefficients, report_type=report_type)

    # ignore the type signature because we override kwargs by specifying the name of
    # the parameters for the user.
    def set_style(  # type: ignore[override]
        self,
        *,
        policy: Literal["override", "update"] = "update",
        barplot_kwargs: dict[str, Any] | None = None,
        boxplot_kwargs: dict[str, Any] | None = None,
        stripplot_kwargs: dict[str, Any] | None = None,
    ):
        """Set the style parameters for the display.

        Parameters
        ----------
        policy : Literal["override", "update"], default="update"
            Policy to use when setting the style parameters.
            If "override", existing settings are set to the provided values.
            If "update", existing settings are not changed; only settings that were
            previously unset are changed.

        barplot_kwargs : dict, default=None
            Keyword arguments to be passed to :func:`seaborn.barplot` for
            rendering the coefficients with an :class:`~skore.EstimatorReport` or
            :class:`~skore.ComparisonReport` of :class:`~skore.EstimatorReport`.

        boxplot_kwargs : dict, default=None
            Keyword arguments to be passed to :func:`seaborn.boxplot` for
            rendering the coefficients with a :class:`~skore.CrossValidationReport` or
            :class:`~skore.ComparisonReport` of :class:`~skore.CrossValidationReport`.

        stripplot_kwargs : dict, default=None
            Keyword arguments to be passed to :func:`seaborn.stripplot` for
            rendering the coefficients with a :class:`~skore.CrossValidationReport` or
            :class:`~skore.ComparisonReport` of :class:`~skore.CrossValidationReport`.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If a style parameter is unknown.
        """
        return super().set_style(
            policy=policy,
            barplot_kwargs=barplot_kwargs or {},
            boxplot_kwargs=boxplot_kwargs or {},
            stripplot_kwargs=stripplot_kwargs or {},
        )



from collections.abc import Sequence
from typing import Literal, cast

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.figure import Figure
from numpy.typing import ArrayLike, NDArray
from sklearn.base import BaseEstimator
from sklearn.metrics import confusion_matrix as sklearn_confusion_matrix
from sklearn.utils._response import _check_response_method

from skore._externals._sklearn_compat import confusion_matrix_at_thresholds
from skore._sklearn._plot.base import DisplayMixin
from skore._sklearn._plot.utils import (
    _ClassifierDisplayMixin,
    _concat_frames_with_column_data,
    _validate_style_kwargs,
)
from skore._sklearn.types import (
    DataSource,
    MLTask,
    PositiveLabel,
    ReportType,
)

ThresholdValue = float | Literal["default"]


class ConfusionMatrixDisplay(_ClassifierDisplayMixin, DisplayMixin):
    """Display for confusion matrix.

    Parameters
    ----------
    confusion_matrix : pd.DataFrame
        Confusion matrix data in long format with columns: "True label",
        "Predicted label", "count", "normalized_by_true", "normalized_by_pred",
        "normalized_by_all" and "threshold". Each row represents one cell of one
        confusion matrix.

    display_labels : list of str
        Display labels for plot axes.

    report_type : {"comparison-cross-validation", "comparison-estimator", \
            "cross-validation", "estimator"}
        The type of report.

    ml_task : {"binary-classification", "multiclass-classification"}
        The machine learning task.

    data_source : {"test", "train"}
        The data source to use.

    pos_label : int, float, bool, str or None
        The class considered as the positive class when displaying the confusion
        matrix.

    response_method : str
        The estimator's method that was used to get the predictions. The possible
        values are: "predict", "predict_proba", and "decision_function".

    Attributes
    ----------
    thresholds : ndarray of shape (n_thresholds,)
        Thresholds of the decision function. Each threshold is associated with a
        confusion matrix. Only available for binary classification. Thresholds are
        sorted in ascending order.
    """

    _default_heatmap_kwargs: dict = {
        "cmap": "Blues",
        "cbar": False,
        "annot": True,
    }

    _default_facet_grid_kwargs: dict = {
        "height": 6,
        "aspect": 1,
    }

    def __init__(
        self,
        *,
        confusion_matrix: pd.DataFrame,
        display_labels: list[str],
        report_type: ReportType,
        ml_task: MLTask,
        thresholds: NDArray,
        data_source: DataSource,
        pos_label: PositiveLabel,
        response_method: str,
    ):
        self.confusion_matrix = confusion_matrix
        self.display_labels = display_labels
        self.report_type = report_type
        self.thresholds = thresholds
        self.ml_task = ml_task
        self.data_source = data_source
        self.pos_label = pos_label
        self.response_method = response_method

    @classmethod
    def _concatenate(
        cls,
        child_displays: Sequence["ConfusionMatrixDisplay"],
        *,
        report_type: ReportType,
        column_data: dict[str, list] | None = None,
        **kwargs,  # for compatibility
    ) -> "ConfusionMatrixDisplay":
        """Build a confusion-matrix display by concatenating child displays."""
        first_display = child_displays[0]
        return cls(
            confusion_matrix=_concat_frames_with_column_data(
                [display.confusion_matrix for display in child_displays],
                column_data,
            ),
            display_labels=first_display.display_labels,
            report_type=report_type,
            ml_task=first_display.ml_task,
            thresholds=np.unique(
                np.concatenate([display.thresholds for display in child_displays])
            ),
            data_source=first_display.data_source,
            pos_label=first_display.pos_label,
            response_method=first_display.response_method,
        )

    @DisplayMixin.style_plot
    def plot(
        self,
        *,
        normalize: Literal["true", "pred", "all"] | None = None,
        threshold_value: ThresholdValue = "default",
        subplot_by: Literal["split", "estimator", "auto"] | None = "auto",
    ) -> Figure:
        """Plot the confusion matrix.

        In binary classification, the confusion matrix can be displayed at various
        decision thresholds. This is useful for understanding how the model's
        predictions change as the decision threshold varies. Use
        ``threshold_value="default"`` to plot at the default threshold (0.5 for
        `predict_proba` response method, 0 for `decision_function` response method).

        Parameters
        ----------
        normalize : {'true', 'pred', 'all'}, default=None
            Normalizes confusion matrix over the true (rows), predicted (columns)
            conditions or all the population. If None, the confusion matrix will not be
            normalized.

        threshold_value : float or "default", default="default"
            The decision threshold to use when applicable (binary classification
            only). If ``"default"``, plots at the default threshold (0.5 for
            `predict_proba` response method, 0 for `decision_function` response
            method). If a float, plots at the closest available threshold.

        subplot_by: Literal["split", "estimator", "auto"] | None = "auto",
            The variable to use for subplotting. If None, the confusion matrix will not
            be subplotted. If "auto", the variable will be automatically determined
            based on the report type.

        Returns
        -------
        matplotlib.figure.Figure
            Figure containing the confusion matrix.
        """
        return self._plot(
            normalize=normalize,
            threshold_value=threshold_value,
            subplot_by=subplot_by,
        )

    def _plot_matplotlib(
        self,
        *,
        normalize: Literal["true", "pred", "all"] | None = None,
        threshold_value: ThresholdValue = "default",
        subplot_by: Literal["split", "estimator", "auto"] | None = "auto",
    ) -> Figure:
        """Matplotlib implementation of the `plot` method.

        Parameters
        ----------
        normalize : {'true', 'pred', 'all'}, default=None
            Normalizes confusion matrix over the true (rows), predicted (columns)
            conditions or all the population. If None, the confusion matrix will not be
            normalized.

        threshold_value : float or "default", default="default"
            The decision threshold to use when applicable.

        subplot_by: Literal["split", "estimator", "auto"] | None = "auto",
            The variable to use for subplotting. If None, the confusion matrix will not
            be subplotted. If "auto", the variable will be automatically determined
            based on the report type.

        Returns
        -------
        matplotlib.figure.Figure
            Figure containing the confusion matrix.
        """
        subplot_by_validated = self._validate_subplot_by(subplot_by, self.report_type)

        if "cross-validation" in self.report_type and subplot_by_validated != "split":
            # Aggregate the data across splits and create custom annotations.
            default_fmt = ".3f" if normalize else ".1f"
            annot_fmt = self._default_heatmap_kwargs.get("fmt", default_fmt)
            frame = self.frame(normalize=normalize, threshold_value=threshold_value)
            aggregated = (
                frame.groupby(
                    ["true_label", "predicted_label", "estimator", "data_source"]
                )["value"]
                .agg(["mean", "std"])
                .reset_index()
            )
            aggregated["annot"] = aggregated.apply(
                lambda row: f"{row['mean']:{annot_fmt}}\n(± {row['std']:{annot_fmt}})",
                axis=1,
            )

            frame = aggregated.rename(columns={"mean": "value"})
            default_fmt = ""
        else:
            frame = self.frame(normalize=normalize, threshold_value=threshold_value)
            default_fmt = ".2f" if normalize else "d"

        heatmap_kwargs_validated = _validate_style_kwargs(
            {"fmt": default_fmt, **self._default_heatmap_kwargs}, {}
        )

        facet_grid_kwargs_validated = _validate_style_kwargs(
            {"col": subplot_by_validated, **self._default_facet_grid_kwargs}, {}
        )
        facet = sns.FacetGrid(
            data=frame,
            **facet_grid_kwargs_validated,
        )
        figure, axes = facet.figure, facet.axes.flatten()

        def plot_heatmap(data, **kwargs):
            """Plot heatmap for each facet."""
            heatmap_data = data.pivot(
                index="true_label", columns="predicted_label", values="value"
            ).reindex(index=self.display_labels, columns=self.display_labels)

            if "cross-validation" in self.report_type and "annot" in data.columns:
                annot_data = data.pivot(
                    index="true_label", columns="predicted_label", values="annot"
                ).reindex(index=self.display_labels, columns=self.display_labels)
                if "annot" in kwargs and kwargs["annot"]:
                    kwargs["annot"] = annot_data
                kwargs["fmt"] = ""

            sns.heatmap(heatmap_data, **kwargs)

        facet.map_dataframe(plot_heatmap, **heatmap_kwargs_validated)

        info_data_source = (
            f"Data source: {self.data_source.capitalize()} set"
            if self.data_source in ("train", "test")
            else None
        )

        title = "Confusion Matrix"
        if self.ml_task == "binary-classification":
            if threshold_value == "default":
                display_threshold = (
                    0.5 if self.response_method == "predict_proba" else 0
                )
            else:
                display_threshold = threshold_value
            title = f"{title}\nDecision threshold: {display_threshold:.2f}"
        figure.suptitle(f"{title}\n{info_data_source}")

        for ax in axes:
            ax.set(
                xlabel="Predicted label",
                ylabel="True label",
            )
            if self.ml_task == "binary-classification" and self.pos_label is not None:
                ticklabels = [
                    f"{label}*" if label == str(self.pos_label) else label
                    for label in self.display_labels
                ]

                ax.set(
                    xticklabels=ticklabels,
                    yticklabels=ticklabels,
                )

                ax.text(
                    -0.15,
                    -0.15,
                    "*: the positive class",
                    fontsize=9,
                    style="italic",
                    verticalalignment="bottom",
                    horizontalalignment="left",
                    transform=ax.transAxes,
                    bbox={
                        "boxstyle": "round",
                        "facecolor": "white",
                        "alpha": 0.8,
                        "edgecolor": "gray",
                    },
                )

        return figure

    def _validate_subplot_by(
        self,
        subplot_by: Literal["split", "estimator", "auto"] | None,
        report_type: ReportType,
    ) -> Literal["split", "estimator"] | None:
        """Validate the `subplot_by` parameter.

        Parameters
        ----------
        subplot_by : Literal["split", "estimator", "auto"] | None
            The variable to use for subplotting.

        report_type : {"comparison-cross-validation", "comparison-estimator", \
                "cross-validation", "estimator"}
            The type of report.

        Returns
        -------
        Literal["split", "estimator"] | None
            The validated `subplot_by` parameter.
        """
        if subplot_by == "auto":
            if "comparison" in report_type:
                return "estimator"
            else:
                return None

        valid_subplot_by: list[Literal["split", "estimator"] | None]
        match report_type:
            case "estimator":
                valid_subplot_by = [None]
            case "cross-validation":
                valid_subplot_by = [None, "split"]
            case "comparison-estimator" | "comparison-cross-validation":
                valid_subplot_by = ["estimator"]

        if subplot_by not in valid_subplot_by:
            raise ValueError(
                f"Invalid `subplot_by` parameter. Valid options are: "
                f"{', '.join(str(s) for s in valid_subplot_by)} or auto. "
                f"Got '{subplot_by}' instead."
            )

        return subplot_by

    @classmethod
    def _compute_data_for_display(
        cls,
        y_true: NDArray | ArrayLike,
        y_pred: NDArray | ArrayLike,
        *,
        report_type: ReportType,
        estimator: BaseEstimator,
        estimator_name: str,
        ml_task: MLTask,
        data_source: DataSource | Literal["both"],
        pos_label: PositiveLabel | None,
        response_method: str | list[str] | tuple[str, ...],
        **kwargs,
    ) -> "ConfusionMatrixDisplay":
        """Compute the confusion matrix for display.

        Parameters
        ----------
        y_true : array-like of shape (n_samples,)
            True labels.

        y_pred : array-like of shape (n_samples,) or (n_samples, n_classes)
            Decision scores when binary classification with thresholds enabled.
            Otherwise, predicted labels.

        report_type : {"comparison-cross-validation", "comparison-estimator", \
                "cross-validation", "estimator"}
            The type of report.

        estimator : BaseEstimator
            The estimator.

        estimator_name : str
            The estimator name to attach to the display data.

        ml_task : {"binary-classification", "multiclass-classification"}
            The machine learning task.

        data_source : {"test", "train"}
            The data source to use.

        pos_label : int, float, bool, str or None
            The class considered as the positive class when displaying the confusion
            matrix.

        response_method : str or list of str or tuple of str
            The estimator's method to be invoked to get the predictions. The possible
            values are: `predict`, `predict_proba`, `predict_log_proba`, and
            `decision_function`.

        **kwargs : dict
            Additional keyword arguments that are ignored for compatibility with
            other metrics displays. Accepts but ignores `estimators`.

        Returns
        -------
        display : ConfusionMatrixDisplay
            The confusion matrix display.
        """
        if data_source == "both":
            raise NotImplementedError(
                "Displaying both data sources is not supported yet."
            )
        data_source = cast(DataSource, data_source)

        classes = estimator.classes_

        if ml_task == "binary-classification":
            # When provided, the positive label is set in second position (which
            # means true-positive counts is the bottom-right cell in the matrix).
            # Usually, TP is the top-left cell, but we align with sklearn.
            if pos_label == classes[0]:
                classes = (classes[1], classes[0])
            tns, fps, fns, tps, thresholds = confusion_matrix_at_thresholds(
                y_true=y_true,
                y_score=y_pred,
                pos_label=classes[1],
            )
            cms = np.column_stack([tns, fps, fns, tps]).reshape(-1, 2, 2).astype(int)
        else:
            cms = sklearn_confusion_matrix(
                y_true=y_true,
                y_pred=y_pred,
                normalize=None,  # we will normalize later
                labels=classes,
            )[np.newaxis, ...]
            thresholds = np.array([np.nan])

        counts = cms.reshape(-1)

        row_sums = cms.sum(axis=2, keepdims=True)
        cm_true = np.zeros_like(cms, dtype=float)
        np.divide(cms, row_sums, out=cm_true, where=row_sums != 0)
        normalized_true_values = cm_true.reshape(-1)

        col_sums = cms.sum(axis=1, keepdims=True)
        cm_pred = np.zeros_like(cms, dtype=float)
        np.divide(cms, col_sums, out=cm_pred, where=col_sums != 0)
        normalized_pred_values = cm_pred.reshape(-1)

        total_sums = cms.sum(axis=(1, 2), keepdims=True)
        cm_all = np.zeros_like(cms, dtype=float)
        np.divide(cms, total_sums, out=cm_all, where=total_sums != 0)
        normalized_all_values = cm_all.reshape(-1)

        n_thresholds = len(thresholds)
        n_classes = len(classes)
        n_cells = n_classes * n_classes

        display_labels = [str(label) for label in classes]
        true_labels = np.tile(np.repeat(display_labels, n_classes), n_thresholds)
        pred_labels = np.tile(np.tile(display_labels, n_classes), n_thresholds)
        threshold_values = np.repeat(thresholds, n_cells)

        confusion_matrix = pd.DataFrame(
            {
                "true_label": true_labels,
                "predicted_label": pred_labels,
                "count": counts,
                "normalized_by_true": normalized_true_values,
                "normalized_by_pred": normalized_pred_values,
                "normalized_by_all": normalized_all_values,
                "threshold": threshold_values,
                "split": None,
                "estimator": estimator_name,
                "data_source": data_source,
            }
        )
        disp = cls(
            confusion_matrix=confusion_matrix,
            display_labels=display_labels,
            report_type=report_type,
            ml_task=ml_task,
            data_source=data_source,
            pos_label=pos_label,
            response_method=_check_response_method(estimator, response_method).__name__,
            thresholds=np.unique(confusion_matrix["threshold"]),
        )

        return disp

    @staticmethod
    def _format_frame(
        df: pd.DataFrame, columns: list[str], normalize_col: str
    ) -> pd.DataFrame:
        return df[columns].rename(columns={normalize_col: "value"})

    def frame(
        self,
        *,
        normalize: Literal["true", "pred", "all"] | None = None,
        threshold_value: ThresholdValue | Literal["all"] = "default",
    ):
        """Return the confusion matrix as a long format dataframe.

        In binary classification, the confusion matrix can be returned at various
        decision thresholds. This is useful for understanding how the model's
        predictions change as the decision threshold varies. Use
        ``threshold_value="default"`` to return the confusion matrix at the default
        threshold (0.5 for `predict_proba` response method, 0 for `decision_function`
        response method). Use ``threshold_value="all"`` to return all available
        thresholds without filtering.

        The matrix is returned as a long format dataframe where each line represents one
        cell of the matrix. The columns are "true_label", "predicted_label", "value",
        "threshold", "split", "estimator", "data_source" ; where "value" is one of
        {"count", "normalized_by_true", "normalized_by_pred", "normalized_by_all"},
        depending on the value of the `normalize` parameter.

        Parameters
        ----------
        normalize : {'true', 'pred', 'all'}, default=None
            Normalizes confusion matrix over the true (rows), predicted (columns)
            conditions or all the population. If None, the confusion matrix will not be
            normalized.

        threshold_value : float, "default" or "all", default="default"
            The decision threshold(s) to use when applicable (binary classification
            only). If ``"default"``, returns the confusion matrix at the default
            threshold (0.5 for `predict_proba` response method, 0 for
            `decision_function` response method). If ``"all"``, returns all available
            thresholds without filtering. If a float, returns the confusion matrix at
            the closest available threshold to the given value.

        Returns
        -------
        frame : pandas.DataFrame
            The confusion matrix as a dataframe.
        """
        normalize_col = "normalized_by_" + normalize if normalize else "count"
        if (
            threshold_value not in ("default", "all")
            and self.ml_task != "binary-classification"
        ):
            raise ValueError(
                "Threshold support is only available for binary classification."
            )

        columns = [
            "true_label",
            "predicted_label",
            normalize_col,
            "threshold",
            "split",
            "estimator",
            "data_source",
        ]

        if threshold_value == "all":
            # Return all thresholds (binary) or full matrix (multiclass).
            return self._format_frame(self.confusion_matrix, columns, normalize_col)
        if threshold_value == "default":
            if self.ml_task == "binary-classification":
                threshold_value = 0.5 if self.response_method == "predict_proba" else 0
            else:
                return self._format_frame(self.confusion_matrix, columns, normalize_col)

        def select_threshold_and_format(group):
            thresholds = np.sort(group["threshold"].unique())
            index_right = int(np.searchsorted(thresholds, threshold_value))
            if index_right == len(thresholds):
                index_right = index_right - 1
            elif index_right == 0 and len(thresholds) > 1:
                index_right = 1
            index_left = index_right - 1
            diff_right = abs(thresholds[index_right] - threshold_value)
            diff_left = abs(thresholds[index_left] - threshold_value)
            closest_threshold_value = thresholds[
                index_right if diff_right < diff_left else index_left
            ]
            frame = group.query(f"threshold == {closest_threshold_value}")
            return self._format_frame(frame, columns, normalize_col)

        frames = []
        if self.report_type == "comparison-cross-validation":
            for _, group in self.confusion_matrix.groupby(
                ["split", "estimator"], observed=True
            ):
                frames.append(select_threshold_and_format(group))
        elif self.report_type == "cross-validation":
            for _, group in self.confusion_matrix.groupby(["split"], observed=True):
                frames.append(select_threshold_and_format(group))
        elif self.report_type == "comparison-estimator":
            for _, group in self.confusion_matrix.groupby(["estimator"], observed=True):
                frames.append(select_threshold_and_format(group))
        else:
            frames.append(select_threshold_and_format(self.confusion_matrix))

        return pd.concat(frames)

    # ignore the type signature because we override kwargs by specifying the name of
    # the parameters for the user.
    def set_style(  # type: ignore[override]
        self,
        *,
        policy: Literal["override", "update"] = "update",
        heatmap_kwargs: dict | None = None,
        facet_grid_kwargs: dict | None = None,
    ):
        """Set the style parameters for the display.

        Parameters
        ----------
        policy : Literal["override", "update"], default="update"
            Policy to use when setting the style parameters.
            If "override", existing settings are set to the provided values.
            If "update", existing settings are not changed; only settings that were
            previously unset are changed.

        heatmap_kwargs : dict, default=None
            Additional keyword arguments to be passed to :func:`seaborn.heatmap`.

        facet_grid_kwargs : dict, default=None
            Additional keyword arguments to be passed to :class:`seaborn.FacetGrid`.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If a style parameter is unknown.
        """
        return super().set_style(
            policy=policy,
            heatmap_kwargs=heatmap_kwargs or {},
            facet_grid_kwargs=facet_grid_kwargs or {},
        )



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Literal

from sklearn.pipeline import Pipeline

from skore._sklearn.types import DataSource


def _check_all_checks(checks: list[Callable]) -> Callable:
    def check(accessor: Any) -> bool:
        return all(check(accessor) for check in checks)

    return check


def _check_has_coef(parent_estimator) -> bool:
    """Check if the estimator or its regressor_ has a specific attribute.

    This is a generic helper function. Please use the appropriate check for your report
    type.
    """
    estimator = (
        parent_estimator.steps[-1][1]
        if isinstance(parent_estimator, Pipeline)
        else parent_estimator
    )
    if hasattr(estimator, "coef_"):
        return True
    try:  # e.g. TransformedTargetRegressor()
        if hasattr(estimator.regressor_, "coef_"):
            return True
    except AttributeError as msg:
        if "object has no attribute 'regressor_'" not in str(msg):
            raise
    raise AttributeError(
        f"Estimator '{estimator}' is not a supported estimator by the function called."
    )


def _check_has_feature_importances(parent_estimator) -> bool:
    """Check if the estimator has a feature_importances_ attribute.

    This is a generic helper function. Please use the appropriate check for your report
    type.
    """
    estimator = (
        parent_estimator.steps[-1][1]
        if isinstance(parent_estimator, Pipeline)
        else parent_estimator
    )
    if hasattr(estimator, "feature_importances_"):
        return True
    raise AttributeError(
        f"Estimator '{estimator}' is not a supported estimator by the function called."
    )


def _check_roc_auc(ml_task_and_methods: list[tuple[str, list[str]]]):
    def check(accessor: Any) -> bool:
        are_supported_cases = []
        for ml_task, methods in ml_task_and_methods:
            is_supported_ml_task = ml_task in accessor._parent._ml_task
            has_methods = any(
                hasattr(accessor._parent._estimator, method) for method in methods
            )
            are_supported_cases.append(is_supported_ml_task and has_methods)

        if not any(are_supported_cases):
            err_msg = (
                f"For the task {accessor._parent._ml_task}, the estimator does not "
                "provide the right prediction methods. The called function requires "
                "the following combinations:\n\n"
            )
            err_msg += "\n".join(
                f"- {ml_task} with {methods}"
                for ml_task, methods in ml_task_and_methods
            )

            raise AttributeError(err_msg)

        return True

    return check


def _expand_data_sources(
    data_source: DataSource | Literal["both"],
) -> tuple[DataSource, ...]:
    """Expand 'both' data source to ('train', 'test')."""
    if data_source == "both":
        return ("train", "test")
    return (data_source,)


########################################################################################
# Accessor related to `EstimatorReport`
########################################################################################


def _check_supported_ml_task(supported_ml_tasks: list[str]) -> Callable:
    def check(accessor: Any) -> bool:
        supported_task = any(
            task in accessor._parent._ml_task for task in supported_ml_tasks
        )

        if not supported_task:
            raise AttributeError(
                f"The {accessor._parent._ml_task} task is not a supported task by "
                f"function called. The supported tasks are {supported_ml_tasks}."
            )

        return True

    return check


def _check_estimator_has_method(method_name: str) -> Callable:
    def check(accessor: Any) -> bool:
        parent_estimator = accessor._parent.estimator_

        if hasattr(parent_estimator, method_name):
            return True

        raise AttributeError(
            f"Estimator '{parent_estimator}' is not a supported estimator by "
            f"the function called. The estimator should have a `{method_name}` "
            "method."
        )

    return check


def _check_estimator_has_coef() -> Callable:
    def check(accessor: Any) -> bool:
        """Check if the estimator has a `coef_` attribute."""
        return _check_has_coef(accessor._parent.estimator_)

    return check


def _check_estimator_has_feature_importances() -> Callable:
    def check(accessor: Any) -> bool:
        """Check if the estimator has a `feature_importances_` attribute."""
        return _check_has_feature_importances(accessor._parent.estimator_)

    return check


########################################################################################
# Accessor related to `CrossValidationReport`
########################################################################################


def _check_estimator_report_has_method(
    accessor_name: str,
    method_name: str,
) -> Callable:
    def check(accessor: Any) -> bool:
        estimator_report = accessor._parent.estimator_reports_[0]

        if not hasattr(estimator_report, accessor_name):
            raise AttributeError(
                f"Estimator report '{estimator_report}' does not have the "
                f"'{accessor_name}' accessor."
            )
        accessor = getattr(estimator_report, accessor_name)

        if hasattr(accessor, method_name):
            return True
        raise AttributeError(
            f"Estimator report '{estimator_report}' is not a supported estimator "
            "by the function called. The estimator report should have a report "
            f"`{method_name}` method."
        )

    return check


def _check_cross_validation_sub_estimator_has_coef() -> Callable:
    def check(accessor: Any) -> bool:
        """Check if the underlying estimator has a `coef_` attribute."""
        return _check_has_coef(accessor._parent.estimator_reports_[0].estimator)

    return check


def _check_cross_validation_sub_estimator_has_feature_importances() -> Callable:
    def check(accessor: Any) -> bool:
        """Check if the underlying estimator has a `feature_importances_` attribute."""
        return _check_has_feature_importances(
            accessor._parent.estimator_reports_[0].estimator_
        )

    return check


########################################################################################
# Accessor related to `ComparisonReport`
########################################################################################


def _check_comparison_report_sub_estimators_have_coef() -> Callable:
    def check(accessor: Any) -> bool:
        """Check if all the estimators have a `coef_` attribute."""
        parent = accessor._parent
        if parent._report_type == "comparison-cross-validation":
            parent_estimators = [
                parent_report.estimator_reports_[0].estimator_
                for parent_report in parent.reports_.values()
            ]
        elif parent._report_type == "comparison-estimator":
            parent_estimators = [
                parent_report.estimator_ for parent_report in parent.reports_.values()
            ]
        else:
            raise TypeError(f"Unexpected report type: {type(parent.reports_[0])}")
        return all(_check_has_coef(e) for e in parent_estimators)

    return check


def _check_comparison_report_sub_estimators_have_feature_importances() -> Callable:
    def check(accessor: Any) -> bool:
        """Check if all the estimators have a `feature_importances_` attribute."""
        parent = accessor._parent
        if parent._report_type == "comparison-cross-validation":
            parent_estimators = [
                parent_report.estimator_reports_[0].estimator_
                for parent_report in parent.reports_.values()
            ]
        elif parent._report_type == "comparison-estimator":
            parent_estimators = [
                parent_report.estimator_ for parent_report in parent.reports_.values()
            ]
        else:
            raise TypeError(f"Unexpected report type: {type(parent.reports_[0])}")
        return all(_check_has_feature_importances(e) for e in parent_estimators)

    return check


########################################################################################
# Accessor related to `ComparisonReport`
########################################################################################


def _check_any_sub_report_has_metric(metric: str) -> Callable[[Any], bool]:
    """Check whether any sub-report of the ComparisonReport supports `metric`."""

    def check(accessor: Any) -> bool:
        return any(
            hasattr(report.metrics, metric)
            for report in accessor._parent.reports_.values()
        )

    return check



from collections.abc import Callable

import pandas as pd
from numpy.typing import ArrayLike
from sklearn.base import BaseEstimator
from sklearn.utils.validation import _num_features


def _function_call_succeeds(func: Callable) -> bool:
    try:
        func()
        return True
    except AttributeError:
        return False


def _get_feature_names(
    estimator: BaseEstimator,
    *,
    transformer: BaseEstimator | None = None,
    X: ArrayLike | pd.DataFrame | None = None,
    n_features: int | None = None,
) -> list[str]:
    """Get the names of an estimator's input features.

    The estimator may or may not be inside a sklearn.Pipeline.
    """
    if hasattr(estimator, "feature_names_in_"):
        feature_names = estimator.feature_names_in_
        # skrub is not implementing rigorously the scikit-learn API and returns a
        # list and so we need to check if we have an array-like before to make any
        # conversion
        if hasattr(feature_names, "tolist"):
            feature_names = feature_names.tolist()
        return feature_names

    elif (
        transformer is not None
        and hasattr(transformer, "get_feature_names_out")
        and _function_call_succeeds(transformer.get_feature_names_out)
    ):
        return transformer.get_feature_names_out().tolist()
    elif X is not None:
        if hasattr(X, "columns"):
            return X.columns.tolist()
        else:
            return [f"Feature #{i}" for i in range(_num_features(X))]

    if n_features is None:
        raise ValueError(
            "Feature names cannot be inferred from the estimator or transformer. "
            "At least one of X or n_features must be provided to infer feature names."
        )
    return [f"Feature #{i}" for i in range(n_features)]



from io import StringIO
from typing import Generic, Literal, TypeVar
from uuid import uuid4

from numpy.typing import ArrayLike
from rich.console import Console
from rich.panel import Panel

from skore._utils.repr.base import AccessorHelpMixin, ReportHelpMixin


class _BaseReport(ReportHelpMixin):
    """Base class for all reports.

    This class centralizes shared report logic (e.g. configuration, accessors) and
    inherits from ``ReportHelpMixin`` to provide a consistent ``help()`` and rich/HTML
    representation across all report types.
    """

    _ACCESSOR_CONFIG: dict[str, dict[str, str]]
    _report_type: Literal[
        "estimator",
        "cross-validation",
        "comparison-estimator",
        "comparison-cross-validation",
    ]

    def __init__(self) -> None:
        self.id = uuid4().int

    @property
    def _hash(self) -> int:
        # FIXME: only for backward compatibility
        return self.id


ParentT = TypeVar("ParentT", bound="_BaseReport")


class _BaseAccessor(AccessorHelpMixin, Generic[ParentT]):
    """Base class for all accessors.

    Accessors expose additional views on a report (e.g. data, metrics) and inherit from
    ``AccessorHelpMixin`` to provide a dedicated ``help()`` and rich/HTML help tree.
    """

    def __init__(self, parent: ParentT) -> None:
        self._parent = parent

    def _rich_repr(self, class_name: str) -> str:
        """Return a string representation using rich for accessors."""
        string_buffer = StringIO()
        console = Console(file=string_buffer, force_terminal=False)
        console.print(
            Panel(
                "Get guidance using the help() method",
                title=f"[cyan]{class_name}[/cyan]",
                border_style="orange1",
                expand=False,
            )
        )
        return string_buffer.getvalue()

    def _get_X_y(
        self,
        *,
        data_source: Literal["test", "train"],
    ) -> tuple[ArrayLike, ArrayLike]:
        """Get the requested dataset.

        Parameters
        ----------
        data_source : {"test", "train"}, default="test"
            The data source to use.

            - "test" : use the test set provided when creating the report.
            - "train" : use the train set provided when creating the report.

        Returns
        -------
        X : array-like of shape (n_samples, n_features)
            The requested dataset.

        y : array-like of shape (n_samples,)
            The requested dataset.
        """
        if data_source == "test":
            if self._parent._X_test is None or self._parent._y_test is None:
                missing_data = "X_test and y_test"
                raise ValueError(
                    f"No {data_source} data (i.e. {missing_data}) were provided "
                    f"when creating the report. Please provide the {data_source} "
                    "data when creating the report."
                )
            return self._parent._X_test, self._parent._y_test
        elif data_source == "train":
            if self._parent._X_train is None or self._parent._y_train is None:
                missing_data = "X_train and y_train"
                raise ValueError(
                    f"No {data_source} data (i.e. {missing_data}) were provided "
                    f"when creating the report. Please provide the {data_source} "
                    "data when creating the report."
                )
            return self._parent._X_train, self._parent._y_train
        else:
            raise ValueError(
                f"Invalid data source: {data_source}. Possible values are: test, train."
            )



import numbers
from collections import namedtuple
from typing import Literal, cast

import numpy as np
import seaborn as sns
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from numpy.typing import ArrayLike
from pandas import DataFrame, Series
from sklearn.utils.validation import _num_samples, check_array

from skore._externals._sklearn_compat import _safe_indexing
from skore._sklearn._plot.base import DisplayMixin
from skore._sklearn._plot.utils import (
    _concat_frames_with_column_data,
    _despine_matplotlib_axis,
    _validate_style_kwargs,
)
from skore._sklearn.types import DataSource, MLTask, ReportType

RangeData = namedtuple("RangeData", ["min", "max"])

MAX_N_LABELS = 6  # 5 + 1 for the perfect model line


class PredictionErrorDisplay(DisplayMixin):
    """Visualization of the prediction error of a regression model.

    This tool can display "residuals vs predicted" or "actual vs predicted"
    using scatter plots to qualitatively assess the behavior of a regressor,
    preferably on held-out data points.

    An instance of this class should be created by
    `EstimatorReport.metrics.prediction_error()`.
    You should not create an instance of this class directly.

    Parameters
    ----------
    prediction_error : DataFrame
        The prediction error data to display. The columns are

        - `estimator`
        - `data_source`
        - `split` (may be null)
        - `output`
        - `y_true`
        - `y_pred`
        - `residuals`.

    range_y_true : RangeData
        Global range of the true values.

    range_y_pred : RangeData
        Global range of the predicted values.

    range_residuals : RangeData
        Global range of the residuals.

    data_source : {"train", "test", "both"}
        The data source used to display the prediction error.

    ml_task : {"regression", "multioutput-regression"}
        The machine learning task.

    report_type : {"comparison-cross-validation", "comparison-estimator", \
            "cross-validation", "estimator"}
        The type of report.

    Examples
    --------
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.linear_model import Ridge
    >>> from skore import evaluate
    >>> X, y = load_diabetes(return_X_y=True)
    >>> classifier = Ridge()
    >>> report = evaluate(classifier, X, y, splitter=0.2)
    >>> display = report.metrics.prediction_error()
    >>> display.plot(kind="actual_vs_predicted")
    """

    _default_relplot_kwargs = {
        "alpha": 0.5,
        "s": 15,
        "marker": "o",
        "aspect": 1.0,
    }
    _default_perfect_model_kwargs = {
        "color": "black",
        "alpha": 0.7,
        "linestyle": "--",
    }

    def __init__(
        self,
        *,
        prediction_error: DataFrame,
        range_y_true: RangeData,
        range_y_pred: RangeData,
        range_residuals: RangeData,
        data_source: DataSource | Literal["both"],
        ml_task: MLTask,
        report_type: ReportType,
    ) -> None:
        self._prediction_error = prediction_error
        self.range_y_true = range_y_true
        self.range_y_pred = range_y_pred
        self.range_residuals = range_residuals
        self.data_source = data_source
        self.ml_task = ml_task
        self.report_type = report_type

    @classmethod
    def _concatenate(
        cls,
        child_displays: list["PredictionErrorDisplay"],
        *,
        report_type: ReportType,
        data_source: None | Literal["both"] = None,
        column_data: dict[str, list] | None = None,
    ) -> "PredictionErrorDisplay":
        """Build a prediction-error display by concatenating child displays."""
        first_display = child_displays[0]
        return cls(
            prediction_error=_concat_frames_with_column_data(
                [display._prediction_error for display in child_displays],
                column_data,
            ),
            range_y_true=RangeData(
                min(display.range_y_true.min for display in child_displays),
                max(display.range_y_true.max for display in child_displays),
            ),
            range_y_pred=RangeData(
                min(display.range_y_pred.min for display in child_displays),
                max(display.range_y_pred.max for display in child_displays),
            ),
            range_residuals=RangeData(
                min(display.range_residuals.min for display in child_displays),
                max(display.range_residuals.max for display in child_displays),
            ),
            data_source=data_source or first_display.data_source,
            ml_task=first_display.ml_task,
            report_type=report_type,
        )

    @DisplayMixin.style_plot
    def plot(
        self,
        *,
        subplot_by: Literal["auto", "data_source", "split", "estimator", "output"]
        | None = "auto",
        kind: Literal[
            "actual_vs_predicted", "residual_vs_predicted"
        ] = "residual_vs_predicted",
        despine: bool = True,
    ) -> Figure:
        """Plot visualization.

        Extra keyword arguments will be passed to matplotlib's ``plot``.

        Parameters
        ----------
        subplot_by : {"auto", "data_source", "split", "estimator", "output"} or \
                None, default="auto"
            The variable to use for creating subplots:

            - "auto" creates subplots by estimator for comparison reports, otherwise
              uses a single plot.
            - "data_source" creates subplots by data source (train/test).
            - "split" creates subplots by cross-validation split.
            - "estimator" creates subplots by estimator.
            - "output" creates subplots by output target.
            - None creates a single plot.

        kind : {"actual_vs_predicted", "residual_vs_predicted"}, \
                default="residual_vs_predicted"
            The type of plot to draw:

            - "actual_vs_predicted" draws the observed values (y-axis) vs.
              the predicted values (x-axis).
            - "residual_vs_predicted" draws the residuals, i.e. difference
              between observed and predicted values, (y-axis) vs. the predicted
              values (x-axis).

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        Returns
        -------
        matplotlib.figure.Figure
            Figure containing the prediction error plot.

        Examples
        --------
        >>> from sklearn.datasets import load_diabetes
        >>> from sklearn.linear_model import Ridge
        >>> from skore import evaluate
        >>> X, y = load_diabetes(return_X_y=True)
        >>> classifier = Ridge()
        >>> report = evaluate(classifier, X, y, splitter=0.2)
        >>> display = report.metrics.prediction_error()
        >>> display.plot(kind="actual_vs_predicted")
        """
        return self._plot(
            subplot_by=subplot_by,
            kind=kind,
            despine=despine,
        )

    def _plot_matplotlib(
        self,
        *,
        subplot_by: Literal["auto", "data_source", "split", "estimator", "output"]
        | None = "auto",
        kind: Literal[
            "actual_vs_predicted", "residual_vs_predicted"
        ] = "residual_vs_predicted",
        despine: bool = True,
    ) -> Figure:
        """Matplolib implementation of the `plot` method."""
        expected_kind = ("actual_vs_predicted", "residual_vs_predicted")
        if kind not in expected_kind:
            raise ValueError(
                f"`kind` must be one of {', '.join(expected_kind)}. "
                f"Got {kind!r} instead."
            )

        if kind == "actual_vs_predicted":
            xlabel, ylabel = "Predicted values", "Actual values"
            y_plot = "y_true"
            min_value = min(self.range_y_pred.min, self.range_y_true.min)
            max_value = max(self.range_y_pred.max, self.range_y_true.max)
            x_range_perfect_pred = [min_value - 0.1, max_value + 0.1]
            y_range_perfect_pred = [min_value - 0.1, max_value + 0.1]
            y_line = y_range_perfect_pred
        else:  # kind == "residual_vs_predicted"
            xlabel, ylabel = "Predicted values", "Residuals (actual - predicted)"
            y_plot = "residuals"
            x_range_perfect_pred = [
                self.range_y_pred.min - 0.1,
                self.range_y_pred.max + 0.1,
            ]
            y_range_perfect_pred = [
                self.range_residuals.min - 0.1,
                self.range_residuals.max + 0.1,
            ]
            y_line = [0, 0]

        plot_data = self.frame()
        col, hue, style = self._get_plot_columns(subplot_by)
        relplot_kwargs = {
            "col": col,
            "hue": hue,
            "style": style,
            **self._default_relplot_kwargs,
        }
        if hue:
            relplot_kwargs["hue_order"] = plot_data[hue].unique().tolist()
        if style == "data_source":
            relplot_kwargs["style_order"] = ["train", "test"]

        facet = sns.relplot(
            data=plot_data,
            x="y_pred",
            y=y_plot,
            kind="scatter",
            **_validate_style_kwargs(relplot_kwargs, {}),
        )
        figure, axes = facet.figure, facet.axes.flatten()

        title = "Prediction Error"
        if "comparison" not in self.report_type:
            title += f" for {self._prediction_error['estimator'].cat.categories.item()}"
        info_data_source = (
            f"Data source: {self.data_source.capitalize()} set"
            if self.data_source in ("train", "test")
            else "Data source: Train and Test"
            if self.data_source == "both"
            else None
        )
        if info_data_source is not None:
            title += f"\n{info_data_source}"
        figure.suptitle(title)

        for ax in axes:
            ax.plot(
                x_range_perfect_pred,
                y_line,
                **self._default_perfect_model_kwargs,
            )
            ax.set(
                xlabel=xlabel,
                ylabel=ylabel,
                xlim=x_range_perfect_pred,
                ylim=y_range_perfect_pred,
                xticks=np.linspace(
                    x_range_perfect_pred[0], x_range_perfect_pred[1], num=5
                ),
                yticks=np.linspace(
                    y_range_perfect_pred[0], y_range_perfect_pred[1], num=5
                ),
            )

            if despine:
                _despine_matplotlib_axis(
                    ax, x_range=ax.get_xlim(), y_range=ax.get_ylim()
                )

        # Add the perfect model line to the legend
        # We retrieve the legend elements created by seaborn, add the perfect model line
        # and create a new legend manually.
        handles = []
        labels = []
        if facet._legend is not None:
            handles = list(facet._legend.legend_handles)
            labels = [t.get_text() for t in facet._legend.get_texts()]
            facet._legend.remove()
            if hue == "split":
                labels = [f"Split #{label}" for label in labels]
            if hue == "output" and style is None:
                labels = [f"Output #{label}" for label in labels]
        handles.append(
            Line2D([0], [0], **self._default_perfect_model_kwargs)  # type: ignore[arg-type]
        )

        labels.append("Perfect predictions")

        axes[len(axes) // 2].legend(
            handles,
            labels,
            loc="upper center",
            bbox_to_anchor=(0.5, -0.3),
            ncol=1,
            frameon=True,
        )

        w, h = figure.get_size_inches()
        figure.set_size_inches(w, h + 0.25 * len(labels))

        return figure

    def _get_plot_columns(
        self,
        subplot_by: Literal["auto", "estimator", "data_source", "split", "output"]
        | None,
    ) -> tuple[str | None, str | None, str | None]:
        """Validate the `subplot_by` parameter.

        Parameters
        ----------
        subplot_by : {"auto", "estimator", "data_source", "split", "output"} or \
                None, default="auto"
            The variable to use for subplotting.

        Returns
        -------
        tuple of (str or None, str or None, str or None)
            A tuple containing (col, hue, style) where:
            - col: Column variable for subplots
            - hue: Variable for color encoding
            - style: Variable for marker style
        """
        is_comparison = "comparison" in self.report_type
        is_crossvalidation = "cross-validation" in self.report_type
        is_multioutput = self.ml_task == "multioutput-regression"
        has_both_sources = self.data_source == "both"
        allow_extra_dims = not (is_multioutput and is_comparison)

        valid_subplot_by: list[str | None] = ["auto"]
        hue_candidates = []

        if has_both_sources and allow_extra_dims:
            valid_subplot_by.append("data_source")
        if is_multioutput:
            valid_subplot_by.append("output")
            hue_candidates.append("output")
        if is_crossvalidation and allow_extra_dims:
            valid_subplot_by.append("split")
            hue_candidates.append("split")
        if is_comparison:
            valid_subplot_by.append("estimator")
            hue_candidates.append("estimator")
        else:
            valid_subplot_by.append(None)

        if subplot_by not in valid_subplot_by:
            raise ValueError(
                f"Invalid `subplot_by` parameter. Valid options are: "
                f"{', '.join(str(s) for s in valid_subplot_by)}. "
                f"Got '{subplot_by}' instead."
            )

        if subplot_by == "auto":
            col = "estimator" if is_comparison else None
        else:
            col = subplot_by
        hue = hue[0] if (hue := [c for c in hue_candidates if c != col]) else None
        style = "data_source" if has_both_sources and col != "data_source" else None

        return col, hue, style

    @classmethod
    def _compute_data_for_display(
        cls,
        y_true: ArrayLike,
        y_pred: ArrayLike,
        *,
        report_type: ReportType,
        estimator_name: str,
        ml_task: MLTask,
        data_source: DataSource,
        subsample: float | int | None = 1_000,
        seed: int | None = None,
        **kwargs,
    ) -> "PredictionErrorDisplay":
        """Plot the prediction error given the true and predicted targets.

        Parameters
        ----------
        y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
            True target values.

        y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
            Predicted target values.

        estimator_name : str
            The estimator name to attach to the display data.

        report_type : {"comparison-cross-validation", "comparison-estimator", \
                "cross-validation", "estimator"}
            The type of report.

        estimators : list of estimator instances
            The estimators from which `y_pred` is obtained.

        ml_task : {"regression", "multioutput-regression"}
            The machine learning task.

        data_source : {"train", "test"}
            The data source used to compute the prediction error curve.

        subsample : float, int or None, default=1_000
            Sampling the samples to be shown on the scatter plot. If `float`,
            it should be between 0 and 1 and represents the proportion of the
            original dataset. If `int`, it represents the number of samples
            display on the scatter plot. If `None`, no subsampling will be
            applied. by default, 1000 samples or less will be displayed.

        seed : int, default=None
            The seed used to initialize the random number generator used for the
            subsampling.

        **kwargs : dict
            Additional keyword arguments that are ignored for compatibility with
            other metrics displays. Here, `estimators` is ignored.

        Returns
        -------
        display : PredictionErrorDisplay
        """
        rng = np.random.default_rng(seed)
        if isinstance(subsample, numbers.Integral):
            if subsample <= 0:
                raise ValueError(
                    f"When an integer, subsample={subsample} should be positive."
                )
        elif isinstance(subsample, numbers.Real) and (subsample <= 0 or subsample >= 1):
            raise ValueError(
                f"When a floating-point, subsample={subsample} should be in the "
                "(0, 1) range."
            )

        if ml_task not in ["regression", "multioutput-regression"]:  # pragma: no cover
            raise ValueError(
                "The machine learning task must be 'regression' or"
                f" 'multioutput-regression'. Got {ml_task} instead."
            )

        n_samples = _num_samples(y_true)
        if subsample is None:
            subsample_ = n_samples
        elif isinstance(subsample, numbers.Integral):
            subsample_ = subsample
        else:  # subsample is a float
            subsample_ = int(n_samples * subsample)

        if subsample_ < n_samples:
            indices = rng.choice(np.arange(n_samples), size=subsample_)
            y_true_sample = check_array(
                _safe_indexing(y_true, indices, axis=0), ensure_2d=False
            )
            y_pred_sample = check_array(
                _safe_indexing(y_pred, indices, axis=0), ensure_2d=False
            )
        else:
            y_true_sample = cast(np.typing.NDArray, y_true)
            y_pred_sample = cast(np.typing.NDArray, y_pred)

        residuals_sample = y_true_sample - y_pred_sample
        n = len(y_true_sample)

        def repeat_category(value: object, n_repeats: int) -> object:
            return Series([value], dtype="category").repeat(n_repeats).array

        if ml_task == "multioutput-regression":
            n_outputs = y_true_sample.shape[1]
            prediction_error_data = {
                "output": (
                    Series(np.arange(n_outputs), dtype="category").repeat(n).array
                ),
                "y_true": y_true_sample.reshape(-1, order="F"),
                "y_pred": y_pred_sample.reshape(-1, order="F"),
                "residuals": residuals_sample.reshape(-1, order="F"),
            }
        else:
            prediction_error_data = {
                "output": repeat_category(np.nan, n),
                "y_true": y_true_sample,
                "y_pred": y_pred_sample,
                "residuals": residuals_sample,
            }

        n_display = len(prediction_error_data["y_true"])
        prediction_error = DataFrame(
            {
                "estimator": repeat_category(estimator_name, n_display),
                "data_source": repeat_category(data_source, n_display),
                "split": repeat_category(None, n_display),
                **prediction_error_data,
            }
        )

        range_y_true = RangeData(min=np.min(y_true_sample), max=np.max(y_true_sample))
        range_y_pred = RangeData(min=np.min(y_pred_sample), max=np.max(y_pred_sample))
        range_residuals = RangeData(
            min=np.min(residuals_sample), max=np.max(residuals_sample)
        )

        return cls(
            prediction_error=prediction_error,
            range_y_true=range_y_true,
            range_y_pred=range_y_pred,
            range_residuals=range_residuals,
            data_source=data_source,
            ml_task=ml_task,
            report_type=report_type,
        )

    def frame(self) -> DataFrame:
        """Get the data used to create the prediction error plot.

        Returns
        -------
        DataFrame
            A DataFrame containing the prediction error data with columns depending on
            the report type:

            - `estimator`: Name of the estimator (when comparing estimators)
            - `split`: Cross-validation split ID (when doing cross-validation)
            - `data_source` : train or test set (when displaying both data sources)
            - `output`: Index of the output target (for multioutput-regression)
            - `y_true`: True target values
            - `y_pred`: Predicted target values
            - `residuals`: Difference between true and predicted values
              `(y_true - y_pred)`


        Examples
        --------
        >>> from sklearn.datasets import load_diabetes
        >>> from sklearn.linear_model import Ridge
        >>> from skore import evaluate
        >>> X, y = load_diabetes(return_X_y=True)
        >>> reg = Ridge()
        >>> report = evaluate(reg, X, y, splitter=0.2)
        >>> display = report.metrics.prediction_error()
        >>> df = display.frame()
        """
        statistical_columns = ["y_true", "y_pred", "residuals"]

        if self.data_source == "both":
            statistical_columns = ["data_source"] + statistical_columns
        if self.ml_task == "multioutput-regression":
            statistical_columns = ["output"] + statistical_columns

        if self.report_type == "estimator":
            columns = statistical_columns
        elif self.report_type == "cross-validation":
            columns = ["split"] + statistical_columns
        elif self.report_type == "comparison-estimator":
            columns = ["estimator"] + statistical_columns
        else:  # self.report_type == "comparison-cross-validation"
            columns = ["estimator", "split"] + statistical_columns

        return self._prediction_error[columns]

    # ignore the type signature because we override kwargs by specifying the name of
    # the parameters for the user.
    def set_style(  # type: ignore[override]
        self,
        *,
        policy: Literal["override", "update"] = "update",
        relplot_kwargs: dict | None = None,
        perfect_model_kwargs: dict | None = None,
    ):
        """Set the style parameters for the display.

        Parameters
        ----------
        policy : {"override", "update"}, default="update"
            Policy to use when setting the style parameters.
            If "override", existing settings are set to the provided values.
            If "update", existing settings are not changed; only settings that were
            previously unset are changed.

        relplot_kwargs : dict, default=None
            Additional keyword arguments to be passed to :func:`seaborn.relplot` for
            rendering the scatter plot(s). Common options include `palette`, `alpha`,
            `s`, `marker`, etc.

        perfect_model_kwargs : dict, default=None
            Additional keyword arguments to be passed to :func:`matplotlib.pyplot.plot`
            for drawing the perfect prediction line. Common options include `color`,
            `alpha`, `linestyle`, etc.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If a style parameter is unknown.
        """
        return super().set_style(
            policy=policy,
            relplot_kwargs=relplot_kwargs or {},
            perfect_model_kwargs=perfect_model_kwargs or {},
        )
